"""Testing formula 3.19 and 20 of EN 1992-1-1:2004."""

import pytest

from blueprints.codes.eurocode.en_1992_1_1_2004.chapter_3_materials.formula_3_19_20 import Form3Dot19And20EffectivePressureZoneHeight


class TestForm3Dot19And20EffectivePressureZoneHeight:
    """Validation for formula 3.19 and 20 from EN 1992-1-1:2004."""

    def test_evaluation_1(self) -> None:
        """Test the evaluation of the result."""
        # Example values
        f_ck = 18.50  # MPa

        form_3_19_20 = Form3Dot19And20EffectivePressureZoneHeight(f_ck=f_ck)

        # Expected result, manually calculated
        manually_calculated_result = 0.8

        assert form_3_19_20 == pytest.approx(expected=manually_calculated_result, rel=1e-4)

    def test_evaluation_2(self) -> None:
        """Test the evaluation of the result."""
        # Example values
        f_ck = 83.5  # MPa

        form_3_19_20 = Form3Dot19And20EffectivePressureZoneHeight(f_ck=f_ck)

        # Expected result, manually calculated
        manually_calculated_result = 0.71625

        assert form_3_19_20 == pytest.approx(expected=manually_calculated_result, rel=1e-4)

    def test_raise_error_when_f_ck_is_larger_than_90(self) -> None:
        """Test a too large value."""
        # Example values
        f_ck = 105  # MPa

        with pytest.raises(ValueError):
            Form3Dot19And20EffectivePressureZoneHeight(f_ck=f_ck)

    @pytest.mark.parametrize(
        ("representation", "expected"),
        [
            (
                "complete",
                r"\lambda = 0.8 = 0.8 = 0.800",
            ),
            ("short", r"\lambda = 0.800"),
        ],
    )
    def test_latex_below_50(self, representation: str, expected: str) -> None:
        """Test the latex representation of the formula."""
        # Example values
        f_ck = 18.50  # -

        # Object to test
        form_3_19_20_latex = Form3Dot19And20EffectivePressureZoneHeight(f_ck=f_ck).latex()

        actual = {"complete": form_3_19_20_latex.complete, "short": form_3_19_20_latex.short}

        assert actual[representation] == expected, f"{representation} representation failed."

    @pytest.mark.parametrize(
        ("representation", "expected"),
        [
            (
                "complete",
                r"\lambda = 0.8 - (f_{ck} - 50) / 400 = 0.8 - (83.500 - 50) / 400 = 0.716",
            ),
            ("short", r"\lambda = 0.716"),
        ],
    )
    def test_latex_above_50(self, representation: str, expected: str) -> None:
        """Test the latex representation of the formula."""
        # Example values
        f_ck = 83.5  # -

        # Object to test
        form_3_19_20_latex = Form3Dot19And20EffectivePressureZoneHeight(f_ck=f_ck).latex()

        actual = {"complete": form_3_19_20_latex.complete, "short": form_3_19_20_latex.short}

        assert actual[representation] == expected, f"{representation} representation failed."



"""Test for the HEA enum."""

from blueprints.structural_sections.steel.steel_cross_sections.standard_profiles.hea import HEA


class TestHEA:
    """Tests for the HEA enum."""

    def test_enum_values(self) -> None:
        """Test that enum values are correctly defined."""
        assert HEA.HEA100.value == ("HEA100", 96, 100, 5, 8, 12)
        assert HEA.HEA1000.value == ("HEA1000", 990, 300, 16.5, 31, 30)

    def test_enum_membership(self) -> None:
        """Test that specific values are members of the enum."""
        assert "HEA100" in [e.value[0] for e in HEA]
        assert "HEA1000" in [e.value[0] for e in HEA]

    def test_enum_uniqueness(self) -> None:
        """Test that all enum values are unique."""
        values = [e.value for e in HEA]
        assert len(values) == len(set(values))

    def test_enum_attributes(self) -> None:
        """Test that enum attributes are correctly assigned."""
        profile = HEA.HEA100
        assert profile.alias == "HEA100"
        assert profile.top_flange_width == 100
        assert profile.top_flange_thickness == 8
        assert profile.bottom_flange_width == 100
        assert profile.bottom_flange_thickness == 8
        assert profile.total_height == 96
        assert profile.web_thickness == 5
        assert profile.top_radius == 12
        assert profile.bottom_radius == 12



"""Testing formula 3.18 of EN 1992-1-1:2004."""

import pytest

from blueprints.codes.eurocode.en_1992_1_1_2004.chapter_3_materials.formula_3_18 import Form3Dot18CompressiveStressConcrete


class TestForm3Dot18CompressiveStressConcrete:
    """Validation for formula 3.18 from EN 1992-1-1:2004."""

    def test_evaluation(self) -> None:
        """Test the evaluation of the result."""
        # Example values
        f_cd = 18.50  # MPa

        form_3_18 = Form3Dot18CompressiveStressConcrete(f_cd=f_cd)

        # Expected result, manually calculated
        manually_calculated_result = 18.50

        assert form_3_18 == pytest.approx(expected=manually_calculated_result, rel=1e-4)

    def test_raise_error_when_negative_f_cd_is_given(self) -> None:
        """Test a negative value."""
        # Example values
        f_cd = -18.50  # MPa

        with pytest.raises(ValueError):
            Form3Dot18CompressiveStressConcrete(f_cd=f_cd)

    @pytest.mark.parametrize(
        ("representation", "expected"),
        [
            (
                "complete",
                r"\sigma_c = f_{cd} = 18.500 = 18.500",
            ),
            ("short", r"\sigma_c = 18.500"),
        ],
    )
    def test_latex(self, representation: str, expected: str) -> None:
        """Test the latex representation of the formula."""
        # Example values
        f_cd = 18.50  # MPa

        # Object to test
        form_3_18_latex = Form3Dot18CompressiveStressConcrete(f_cd=f_cd).latex()

        actual = {"complete": form_3_18_latex.complete, "short": form_3_18_latex.short}

        assert actual[representation] == expected, f"{representation} representation failed."



"""Formula 3.23 from EN 1992-1-1:2004: Chapter 3 - Materials."""

from blueprints.codes.eurocode.en_1992_1_1_2004 import EN_1992_1_1_2004
from blueprints.codes.formula import Formula
from blueprints.codes.latex_formula import LatexFormula
from blueprints.type_alias import MM, MPA


class Form3Dot23FlexuralTensileStrength(Formula):
    """Class representing formula 3.23 for the calculation of the mean flexural tensile strength of reinforced concrete members."""

    label = "3.23"
    source_document = EN_1992_1_1_2004

    def __init__(
        self,
        h: MM,
        f_ctm: MPA,
    ) -> None:
        r"""[$f_{ctm,fl}$] Mean flexural tensile strength of reinforced concrete members  [$MPa$].

        EN 1992-1-1:2004 art.3.1.8(1) - Formula (3.23)

        Parameters
        ----------
        h : MM
            [${h}$] Total member depth [mm].
        f_ctm : MPA
            [${f_{ctm}}$] Mean axial tensile strength following from table 3.1 [$MPa$].

        Returns
        -------
        None
        """
        super().__init__()
        self.h = h
        self.f_ctm = f_ctm

    @staticmethod
    def _evaluate(
        h: MM,
        f_ctm: MPA,
    ) -> MPA:
        """Evaluates the formula, for more information see the __init__ method."""
        if h < 0:
            raise ValueError(f"Invalid h: {h}. h cannot be negative")
        if f_ctm < 0:
            raise ValueError(f"Invalid f_ctm: {f_ctm}. f_ctm cannot be negative")
        return max((1.6 - h / 1000) * f_ctm, f_ctm)

    def latex(self, n: int = 3) -> LatexFormula:
        """Returns LatexFormula object for formula 3.23."""
        return LatexFormula(
            return_symbol=r"f_{ctm,fl}",
            result=f"{self:.{n}f}",
            equation=r"\max \left[ (1.6 - h/1000) \cdot f_{ctm} ; f_{ctm} \right]",
            numeric_equation=rf"\max \left[ (1.6 - {self.h:.{n}f}/1000) \cdot {self.f_ctm:.{n}f} ; {self.f_ctm:.{n}f} \right]",
            comparison_operator_label="=",
        )



"""Fixtures for structural sections tests."""

import pytest
from sectionproperties.post.post import SectionProperties

from blueprints.structural_sections.cross_section_annular_sector import AnnularSectorCrossSection
from blueprints.structural_sections.cross_section_circle import CircularCrossSection
from blueprints.structural_sections.cross_section_cornered import CircularCorneredCrossSection
from blueprints.structural_sections.cross_section_hexagon import HexagonalCrossSection
from blueprints.structural_sections.cross_section_rectangle import RectangularCrossSection
from blueprints.structural_sections.cross_section_triangle import RightAngledTriangularCrossSection
from blueprints.structural_sections.cross_section_tube import TubeCrossSection


@pytest.fixture
def rectangular_cross_section() -> RectangularCrossSection:
    """Return a RectangularCrossSection instance."""
    return RectangularCrossSection(name="Rectangle", width=100.0, height=200.0, x=100.0, y=250.0)


@pytest.fixture
def rectangular_cross_section_section_properties() -> SectionProperties:
    """Return a RectangularCrossSection instance."""
    rect = RectangularCrossSection(name="Rectangle", width=100.0, height=200.0, x=100.0, y=250.0)
    # Calculate properties
    return rect.section_properties(geometric=True, plastic=True, warping=True)


@pytest.fixture
def circular_cross_section() -> CircularCrossSection:
    """Return a CircularCrossSection instance."""
    return CircularCrossSection(name="Circle", diameter=200.0, x=100.0, y=250.0)


@pytest.fixture
def tube_cross_section() -> TubeCrossSection:
    """Return a TubeCrossSection instance."""
    return TubeCrossSection(name="Tube", outer_diameter=100.0, inner_diameter=50.0, x=100.0, y=250.0)


@pytest.fixture
def triangular_cross_section() -> RightAngledTriangularCrossSection:
    """Return a RightAngledTriangularCrossSection instance."""
    return RightAngledTriangularCrossSection(name="Triangle", base=100.0, height=200.0, x=100.0, y=250.0)


@pytest.fixture
def qcs_cross_section() -> CircularCorneredCrossSection:
    """Return a CircularCorneredCrossSection instance."""
    return CircularCorneredCrossSection(inner_radius=50.0, outer_radius=0, x=100.0, y=250.0, thickness_vertical=0.0, thickness_horizontal=0.0)


@pytest.fixture
def hexagonal_cross_section() -> HexagonalCrossSection:
    """Return a HexagonalCrossSection instance."""
    return HexagonalCrossSection(name="Hexagon", side_length=50.0, x=100.0, y=250.0)


@pytest.fixture
def annular_sector_cross_section() -> AnnularSectorCrossSection:
    """Return an AnnularSectorCrossSection instance."""
    return AnnularSectorCrossSection(
        inner_radius=90.0,
        thickness=20.0,
        start_angle=0.0,
        end_angle=90.0,
        x=100.0,
        y=250.0,
        name="AnnularSector",
    )


@pytest.fixture
def annular_sector_cross_section_359_degrees() -> AnnularSectorCrossSection:
    """Return an AnnularSectorCrossSection instance."""
    return AnnularSectorCrossSection(
        inner_radius=90.0,
        thickness=20.0,
        start_angle=90.0,
        end_angle=90.0 + 359.0,
        x=0.0,
        y=0.0,
        name="AnnularSector",
    )



"""Formula 5.36 from EN 1992-1-1:2004: Chapter 5 - Structural Analysis."""

from blueprints.codes.eurocode.en_1992_1_1_2004 import EN_1992_1_1_2004
from blueprints.codes.formula import Formula
from blueprints.codes.latex_formula import LatexFormula
from blueprints.type_alias import DIMENSIONLESS, KN, MM2, MPA
from blueprints.validations import raise_if_less_or_equal_to_zero, raise_if_negative


class Form5Dot36RelativeAxialForce(Formula):
    r"""Class representing formula 5.36 for the calculation of the relative axial force, [$K_{r}$]."""

    label = "5.36"
    source_document = EN_1992_1_1_2004

    def __init__(
        self,
        n_ed: KN,
        ac: MM2,
        fcd: MPA,
        as_: MM2,
        fyd: MPA,
        n_bal: DIMENSIONLESS,
    ) -> None:
        r"""[$K_{r}$] Relative axial force [-].

        EN 1992-1-1:2004 art.5.8.8.3(3) - Formula (5.36)

        Parameters
        ----------
        n_ed : KN
            [$N_{Ed}$] Design value of axial load [$kN$].
        ac : MM2
            [$A_{c}$] Area of concrete cross section [$mm^2$].
        fcd : MPA
            [$f_{cd}$] Design value of concrete compressive strength [$MPa$].
        as_ : MM2
            [$A_{s}$] Total area of reinforcement [$mm^2$].
        fyd : MPA
            [$f_{yd}$] Design yield strength of reinforcement [$MPa$].
        n_bal : DIMENSIONLESS
            [$n_{bal}$] Value of n at maximum moment resistance, 0.4 may be used [-].
        """
        super().__init__()
        self.n_ed = n_ed
        self.ac = ac
        self.fcd = fcd
        self.as_ = as_
        self.fyd = fyd
        self.n_bal = n_bal

    @staticmethod
    def _evaluate(
        n_ed: KN,
        ac: MM2,
        fcd: MPA,
        as_: MM2,
        fyd: MPA,
        n_bal: float,
    ) -> float:
        """Evaluates the formula, for more information see the __init__ method."""
        raise_if_negative(n_ed=n_ed, as_=as_, fyd=fyd, n_bal=n_bal)
        raise_if_less_or_equal_to_zero(ac=ac, fcd=fcd)

        omega = as_ * fyd / (ac * fcd)
        nu = 1 + omega
        n = n_ed / (ac * fcd)
        return min((nu - n) / (nu - n_bal), 1)

    def latex(self, n: int = 3) -> LatexFormula:
        """Returns LatexFormula object for formula 5.36."""
        return LatexFormula(
            return_symbol=r"K_{r}",
            result=f"{self:.{n}f}",
            equation=r"\min\left(\frac{\left(1 + \frac{A_{s} \cdot f_{yd}}{A_{c} \cdot f_{cd}}\right) - "
            r"\frac{N_{Ed}}{A_{c} \cdot f_{cd}}}{\left(1 + \frac{A_{s} \cdot f_{yd}}{A_{c} \cdot f_{cd}}\right) - n_{bal}}, 1\right)",
            numeric_equation=rf"\min\left(\frac{{\left(1 + \frac{{{self.as_:.{n}f} \cdot {self.fyd:.{n}f}}}{{{self.ac:.{n}f} \cdot "
            rf"{self.fcd:.{n}f}}}\right) - \frac{{{self.n_ed:.{n}f}}}{{{self.ac:.{n}f}"
            rf" \cdot {self.fcd:.{n}f}}}}}{{\left(1 + \frac{{{self.as_:.{n}f} \cdot "
            rf"{self.fyd:.{n}f}}}{{{self.ac:.{n}f} \cdot {self.fcd:.{n}f}}}\right) - {self.n_bal:.{n}f}}}, 1\right)",
            comparison_operator_label="=",
            unit="-",
        )



"""Testing formula 7.20 of EN 1992-1-1:2004."""

import pytest

from blueprints.codes.eurocode.en_1992_1_1_2004.chapter_7_serviceability_limit_state.formula_7_20 import Form7Dot20EffectiveModulusCreep
from blueprints.validations import NegativeValueError


class TestForm7Dot20EffectiveModulusCreep:
    """Validation for formula 7.20 from EN 1992-1-1:2004."""

    def test_evaluation(self) -> None:
        """Tests the evaluation of the result."""
        # Example values
        e_cm = 30000.0  # MPa
        phi_inf_t0 = 2.0  # dimensionless

        # Object to test
        formula = Form7Dot20EffectiveModulusCreep(e_cm=e_cm, phi_inf_t0=phi_inf_t0)

        # Expected result, manually calculated
        manually_calculated_result = 10000.0  # MPa

        assert formula == pytest.approx(manually_calculated_result, rel=1e-4)

    @pytest.mark.parametrize(
        ("e_cm", "phi_inf_t0"),
        [
            (-30000.0, 2.0),  # e_cm is negative
            (30000.0, -2.0),  # phi_inf_t0 is negative
        ],
    )
    def test_raise_error_when_invalid_values_are_given(self, e_cm: float, phi_inf_t0: float) -> None:
        """Test invalid values."""
        with pytest.raises(NegativeValueError):
            Form7Dot20EffectiveModulusCreep(e_cm=e_cm, phi_inf_t0=phi_inf_t0)

    @pytest.mark.parametrize(
        ("representation", "expected"),
        [
            (
                "complete",
                r"E_{c,eff} = \frac{E_{cm}}{1 + \phi(\infty , t_0)} = \frac{30000.000}{1 + 2.000} = 10000.000 \ MPa",
            ),
            ("short", r"E_{c,eff} = 10000.000 \ MPa"),
        ],
    )
    def test_latex(self, representation: str, expected: str) -> None:
        """Test the latex representation of the formula."""
        # Example values
        e_cm = 30000.0  # MPa
        phi_inf_t0 = 2.0  # dimensionless

        # Object to test
        latex = Form7Dot20EffectiveModulusCreep(e_cm=e_cm, phi_inf_t0=phi_inf_t0).latex()

        actual = {
            "complete": latex.complete,
            "short": latex.short,
        }

        assert expected == actual[representation], f"{representation} representation failed."



"""Testing sub-formulas for 8.2 of EN 1992-1-1:2004."""

import pytest

from blueprints.codes.eurocode.en_1992_1_1_2004.chapter_8_detailing_of_reinforcement_and_prestressing_tendons.formula_8_2 import (
    SubForm8Dot2CoefficientBarDiameter,
)
from blueprints.validations import NegativeValueError


class TestSubForm8Dot2CoefficientBarDiameter:
    """Validation for sub-formula 8.2 from EN 1992-1-1:2004."""

    def test_evaluation_small_diameter(self) -> None:
        """Test the evaluation of the result for diameters smaller than 32 mm."""
        diameter = 16  # str
        sub_form_8_2_3 = SubForm8Dot2CoefficientBarDiameter(diameter=diameter)

        # Expected result, manually calculated
        manually_result = 1
        assert sub_form_8_2_3 == pytest.approx(expected=manually_result, rel=1e-4)

    def test_evaluation_large_diameter(self) -> None:
        """Test the evaluation of the result for diameters larger than 32 mm."""
        diameter = 40  # str
        sub_form_8_2_4 = SubForm8Dot2CoefficientBarDiameter(diameter=diameter)

        # Expected result, manually calculated
        manually_result = 0.92
        assert sub_form_8_2_4 == pytest.approx(expected=manually_result, rel=1e-4)

    def test_raise_error_when_negative_diameter_is_given(self) -> None:
        """Test an invalid bond quality."""
        # Example values
        diameter = -1  # mm

        with pytest.raises(NegativeValueError):
            SubForm8Dot2CoefficientBarDiameter(diameter=diameter)

    @pytest.mark.parametrize(
        ("representation", "expected_result"),
        [
            (
                "complete_latex_16",
                r"\eta_2 = \begin{matrix} 1.0 & \text{for }Ø ≤ 32 \\ (132 - Ø) / 100 & \text{for }Ø > 32  \end{matrix} = 1.00 = 1.00",
            ),
            ("short_latex_16", r"\eta_2 = 1.00"),
            (
                "complete_latex_64",
                r"\eta_2 = \begin{matrix} 1.0 & \text{for }Ø ≤ 32 \\ (132 - Ø) / 100 & " r"\text{for }Ø > 32  \end{matrix} = (132 - 64) / 100 = 0.68",
            ),
            ("short_latex_64", r"\eta_2 = 0.68"),
        ],
    )
    def test_latex(self, representation: str, expected_result: str) -> None:
        """Test the latex representation."""
        latex_16 = SubForm8Dot2CoefficientBarDiameter(diameter=16).latex()
        latex_64 = SubForm8Dot2CoefficientBarDiameter(diameter=64).latex()

        actual = {
            "complete_latex_16": latex_16.complete,
            "short_latex_16": latex_16.short,
            "complete_latex_64": latex_64.complete,
            "short_latex_64": latex_64.short,
        }

        assert actual[representation] == expected_result, f"{representation} representation failed."



"""Formula 5.41 from EN 1992-1-1:2004: Chapter 5 - Structural Analysis."""

from blueprints.codes.eurocode.en_1992_1_1_2004 import EN_1992_1_1_2004
from blueprints.codes.formula import Formula
from blueprints.codes.latex_formula import LatexFormula
from blueprints.type_alias import DIMENSIONLESS, MM2, MPA, N
from blueprints.validations import raise_if_negative


class Form5Dot41MaxForceTendon(Formula):
    r"""Class representing formula 5.41 for the calculation of the maximum force applied to a tendon, [$P_{max}$]."""

    label = "5.41"
    source_document = EN_1992_1_1_2004

    def __init__(
        self,
        a_p: MM2,
        k_1: DIMENSIONLESS,
        f_pk: MPA,
        k_2: DIMENSIONLESS,
        f_p0_1k: MPA,
    ) -> None:
        r"""[$P_{max}$] Maximum force applied to a tendon at active end [$N$].

        EN 1992-1-1:2004 art.5.10.2.1(1) - Formula (5.41)

        Parameters
        ----------
        a_p : MM2
            [$A_{p}$] Cross-sectional area of the tendon [$mm^2$].
        k_1 : DIMENSIONLESS
            [$k_{1}$] Coefficient for characteristic tensile strength, recommended value is 0.8 [$-$].
        f_pk : MPA
            [$f_{pk}$] Characteristic tensile strength of the tendon [$MPa$].
        k_2 : DIMENSIONLESS
            [$k_{2}$] Coefficient for 0.1% proof stress, recommended value is 0.9 [$-$].
        f_p0_1k : MPA
            [$f_{p0.1k}$] 0.1% proof stress of the tendon [$MPa$].
        """
        super().__init__()
        self.a_p = a_p
        self.k_1 = k_1
        self.f_pk = f_pk
        self.k_2 = k_2
        self.f_p0_1k = f_p0_1k

    @staticmethod
    def _evaluate(
        a_p: MM2,
        k_1: DIMENSIONLESS,
        f_pk: MPA,
        k_2: DIMENSIONLESS,
        f_p0_1k: MPA,
    ) -> N:
        """Evaluates the formula, for more information see the __init__ method."""
        raise_if_negative(
            a_p=a_p,
            k_1=k_1,
            f_pk=f_pk,
            k_2=k_2,
            f_p0_1k=f_p0_1k,
        )

        return a_p * min(k_1 * f_pk, k_2 * f_p0_1k)

    def latex(self, n: int = 3) -> LatexFormula:
        """Returns LatexFormula object for formula 5.41."""
        return LatexFormula(
            return_symbol=r"P_{max}",
            result=f"{self:.{n}f}",
            equation=r"A_{p} \cdot \min(k_1 \cdot f_{pk}, k_2 \cdot f_{p0.1k})",
            numeric_equation=rf"{self.a_p:.{n}f} \cdot \min({self.k_1:.{n}f} \cdot {self.f_pk:.{n}f}, {self.k_2:.{n}f} \cdot {self.f_p0_1k:.{n}f})",
            comparison_operator_label="=",
            unit="N",
        )



"""Testing nominal concrete cover check of EN 1992-1-1."""

import pytest

from blueprints.checks.nominal_concrete_cover.nominal_concrete_cover import NominalConcreteCover
from blueprints.codes.eurocode.en_1992_1_1_2004.chapter_4_durability_and_cover._base_classes.nominal_cover_constants import (
    AbrasionClass,
    CastingSurface,
    NominalConcreteCoverConstantsBase,
)
from blueprints.codes.eurocode.en_1992_1_1_2004.chapter_4_durability_and_cover._base_classes.structural_class import ConcreteStructuralClassBase
from blueprints.codes.eurocode.en_1992_1_1_2004.chapter_4_durability_and_cover.constants import NominalConcreteCoverConstants
from blueprints.codes.eurocode.en_1992_1_1_2004.chapter_4_durability_and_cover.table_4_1 import (
    Carbonation,
    Chemical,
    Chloride,
    ChlorideSeawater,
    FreezeThaw,
    Table4Dot1ExposureClasses,
)
from blueprints.codes.eurocode.en_1992_1_1_2004.chapter_4_durability_and_cover.table_4_3 import Table4Dot3ConcreteStructuralClass
from blueprints.materials.concrete import ConcreteMaterial
from blueprints.type_alias import MM

exposure_classes = Table4Dot1ExposureClasses(Carbonation.XC1, Chloride.XD1, ChlorideSeawater.XS1, FreezeThaw.NA, Chemical.NA)
structural_class = Table4Dot3ConcreteStructuralClass(exposure_classes, 50, ConcreteMaterial(), False, False)


class TestNominalConcreteCover:
    """Validation for nominal concrete cover check from EN 1992-1-1."""

    @pytest.mark.parametrize(
        (
            "reinforcement_diameter",
            "nominal_max_aggregate_size",
            "constants",
            "structural_class",
            "carbonation",
            "chloride",
            "chloride_seawater",
            "delta_c_dur_gamma",
            "delta_c_dur_st",
            "delta_c_dur_add",
            "casting_surface",
            "uneven_surface",
            "abrasion_class",
            "expected",
        ),
        [
            (
                25,
                False,
                NominalConcreteCoverConstants(),
                structural_class,
                Carbonation.XC1,
                Chloride.XD1,
                ChlorideSeawater.XS1,
                0,
                0,
                0,
                CastingSurface.PERMANENTLY_EXPOSED,
                False,
                AbrasionClass.XM1,
                50,
            ),
            (
                25,
                False,
                NominalConcreteCoverConstants(),
                structural_class,
                "XC1",
                "XD1",
                "XS1",
                0,
                0,
                0,
                CastingSurface.PERMANENTLY_EXPOSED,
                False,
                AbrasionClass.XM1,
                50,
            ),
            (
                32,
                True,
                NominalConcreteCoverConstants(),
                6,
                Carbonation.XC3,
                Chloride.XD2,
                ChlorideSeawater.XS2,
                0,
                0,
                0,
                CastingSurface.PREPARED_GROUND,
                False,
                AbrasionClass.XM2,
                90,
            ),
            (
                20,
                False,
                NominalConcreteCoverConstants(),
                2,
                Carbonation.XC1,
                Chloride.XD3,
                ChlorideSeawater.XS3,
                0,
                0,
                0,
                CastingSurface.DIRECTLY_AGAINST_SOIL,
                False,
                AbrasionClass.XM2,
                105,
            ),
        ],
    )
    def test_evaluation(  # noqa: PLR0913
        self,
        reinforcement_diameter: MM,
        nominal_max_aggregate_size: bool,
        constants: NominalConcreteCoverConstantsBase,
        structural_class: ConcreteStructuralClassBase,
        carbonation: Carbonation,
        chloride: Chloride,
        chloride_seawater: ChlorideSeawater,
        delta_c_dur_gamma: MM,
        delta_c_dur_st: MM,
        delta_c_dur_add: MM,
        casting_surface: CastingSurface,
        uneven_surface: bool,
        abrasion_class: AbrasionClass,
        expected: MM,
    ) -> None:
        """Test the evaluation of the result."""
        nominal_concrete_cover = NominalConcreteCover(
            reinforcement_diameter=reinforcement_diameter,
            nominal_max_aggregate_size=nominal_max_aggregate_size,
            constants=constants,
            structural_class=structural_class,
            carbonation=carbonation,
            chloride=chloride,
            chloride_seawater=chloride_seawater,
            delta_c_dur_gamma=delta_c_dur_gamma,
            delta_c_dur_st=delta_c_dur_st,
            delta_c_dur_add=delta_c_dur_add,
            casting_surface=casting_surface,
            uneven_surface=uneven_surface,
            abrasion_class=abrasion_class,
        ).value()

        assert nominal_concrete_cover == pytest.approx(expected=expected, rel=1e-4)

    @pytest.mark.parametrize(
        "uneven_surface",
        [1, "False", None, 0.5],
    )
    def test_uneven_surface_type_error(
        self,
        uneven_surface: bool,
    ) -> None:
        """Test type error for uneven_surface parameter."""
        with pytest.raises(TypeError, match="Invalid type for uneven_surface: .* Expected type is bool."):
            NominalConcreteCover(
                reinforcement_diameter=25,
                nominal_max_aggregate_size=32,
                constants=NominalConcreteCoverConstants(),
                structural_class=structural_class,
                carbonation=Carbonation.XC1,
                chloride=Chloride.XD1,
                chloride_seawater=ChlorideSeawater.XS1,
                delta_c_dur_gamma=0,
                delta_c_dur_st=0,
                delta_c_dur_add=0,
                casting_surface=CastingSurface.PERMANENTLY_EXPOSED,
                uneven_surface=uneven_surface,
                abrasion_class=AbrasionClass.XM1,
            )

    @pytest.mark.parametrize(
        "abrasion_class",
        [1, "XM1", None, 0.5],
    )
    def test_abrasion_class_type_error(
        self,
        abrasion_class: AbrasionClass,
    ) -> None:
        """Test type error for abrasion_class parameter."""
        with pytest.raises(TypeError, match="Invalid type for abrasion_class: .* Expected type is AbrasionClass."):
            NominalConcreteCover(
                reinforcement_diameter=25,
                nominal_max_aggregate_size=32,
                constants=NominalConcreteCoverConstants(),
                structural_class=structural_class,
                carbonation=Carbonation.XC1,
                chloride=Chloride.XD1,
                chloride_seawater=ChlorideSeawater.XS1,
                delta_c_dur_gamma=0,
                delta_c_dur_st=0,
                delta_c_dur_add=0,
                casting_surface=CastingSurface.PERMANENTLY_EXPOSED,
                uneven_surface=False,
                abrasion_class=abrasion_class,
            )

    @pytest.mark.parametrize(
        "casting_surface",
        [1, "PERMANENTLY_EXPOSED", None, 0.5],
    )
    def test_casting_surface_type_error(
        self,
        casting_surface: CastingSurface,
    ) -> None:
        """Test type error for casting_surface parameter."""
        with pytest.raises(TypeError, match="Invalid type for casting_surface: .* Expected type is CastingSurface."):
            NominalConcreteCover(
                reinforcement_diameter=25,
                nominal_max_aggregate_size=32,
                constants=NominalConcreteCoverConstants(),
                structural_class=structural_class,
                carbonation=Carbonation.XC1,
                chloride=Chloride.XD1,
                chloride_seawater=ChlorideSeawater.XS1,
                delta_c_dur_gamma=0,
                delta_c_dur_st=0,
                delta_c_dur_add=0,
                casting_surface=casting_surface,
                uneven_surface=False,
                abrasion_class=AbrasionClass.XM1,
            )

    def test_latex_representation(self) -> None:
        """Test the latex representation of the formula."""
        nominal_concrete_cover = NominalConcreteCover(
            reinforcement_diameter=25,
            nominal_max_aggregate_size=40,
            constants=NominalConcreteCoverConstants(),
            structural_class=structural_class,
            carbonation=Carbonation.XC1,
            chloride=Chloride.XD1,
            chloride_seawater=ChlorideSeawater.XS1,
            delta_c_dur_gamma=0,
            delta_c_dur_st=0,
            delta_c_dur_add=0,
            casting_surface=CastingSurface.DIRECTLY_AGAINST_SOIL,
            uneven_surface=True,
            abrasion_class=AbrasionClass.XM1,
        )

        assert (
            nominal_concrete_cover.latex() == r"Nominal~concrete~cover~according~to~art.~4.4.1~from~EN~1992-1-1:2004:\newline~"
            r"\max~\left\{Nominal~concrete~cover~according~to~art.~4.4.1~(c_{nom});~Minimum~cover~with~regard~to~casting~surface~according~to~art.~4.4.1.3~(4)\right\}\newline~"
            r"=~\max~\left\{55.0;~110.0\right\}~=~110.0~mm\newline~"
            r"\newline~"
            r"Where:\newline~"
            r"\hspace{4ex}c_{nom}~=~c_{min,total}+\Delta~c_{dev}~=~45.0+10~=~55.0~mm\newline~"
            r"\hspace{4ex}\Delta~c_{dev}~is~determined~according~to~art.~4.4.1.3~(1)\newline~"
            r"\hspace{4ex}c_{min,total}~=~c_{min}~+~\Delta~c_{uneven~surface}~~+~\Delta~c_{abrasion~class}~=~35.0~+~5~+~5~=~45.0~mm\newline~"
            r"\hspace{4ex}\Delta~c_{uneven~surface}~and~\Delta~c_{abrasion~class}~are~determined~according~to~art.~4.4.1.2~(11)~and~(13)\newline~"
            r"\hspace{4ex}c_{min}~=~\max~\left\{c_{min,b};~c_{min,dur}+\Delta~c_{dur,\gamma}-\Delta~c_{dur,st}-\Delta~c_{dur,add};~10~\text{mm}\right\}~=~\max~\left\{30.0;~35.0+0-0-0;~10\right\}~=~35.0~mm\newline~"
            r"\hspace{4ex}\Delta~c_{dur,\gamma}~,~\Delta~c_{dur,st}~and~\Delta~c_{dur,add}~are~determined~according~to~art.~4.4.1.2~(6),~(7)~and~(8)\newline~"
            r"\hspace{4ex}c_{min,b}~is~determined~according~to~table~4.2~based~on~(equivalent)~rebar~diameter~+~5~=~25~+~5~=~30~mm\newline~"
            r"\hspace{4ex}c_{min,dur}~is~determined~according~to~table~4.4~based~on~structural~class~S4~\&~exposure~classes~(XC1,~XD1,~XS1)~=~35~mm\newline~"
            r"\hspace{4ex}Minimum~cover~with~regard~to~casting~surface~according~to~art.~4.4.1.3~(4)~=~k2~\ge~c_{min,dur}~+~75~mm~for~Directly~against~soil"
        )

        assert str(nominal_concrete_cover) == "Nominal concrete cover according to art. 4.4.1 = 110.0 mm"



"""Formula 6.52 from EN 1992-1-1:2004: Chapter 6 - Ultimate Limit State."""

import numpy as np

from blueprints.codes.eurocode.en_1992_1_1_2004 import EN_1992_1_1_2004
from blueprints.codes.formula import Formula
from blueprints.codes.latex_formula import LatexFormula, latex_replace_symbols
from blueprints.type_alias import DEG, MM, MM2, MPA
from blueprints.validations import raise_if_negative


class Form6Dot52PunchingShearResistance(Formula):
    r"""Class representing formula 6.52 for the calculation of punching shear resistance.

    EN 1992-1-1:2004 art.6.4.5(1) - Formula (6.52)

    Parameters
    ----------
    v_rd_c : MPA
        [$v_{Rd,c}$] Design shear strength of concrete without shear reinforcement [$MPa$].
    d : MM
        [$d$] Mean effective depth of the slab [$mm$].
    s_r : MM
        [$s_r$] Radial spacing of perimeters of shear reinforcement [$mm$].
    a_sw : MM2
        [$A_{sw}$] Area of one perimeter of shear reinforcement around the column [$mm^2$].
    f_ywd_ef : MPA
        [$f_{ywd,ef}$] Effective design strength of the punching shear reinforcement [$MPa$].
    u_1 : MM
        [$u_{y1}$] Perimeter of the critical section [$mm$].
    alpha : DEG
        [$\alpha$] Angle between the shear reinforcement and the plane of the slab [$deg$].
    """

    label = "6.52"
    source_document = EN_1992_1_1_2004

    def __init__(
        self,
        v_rd_c: MPA,
        d: MM,
        s_r: MM,
        a_sw: MM2,
        f_ywd_ef: MPA,
        u_1: MM,
        alpha: DEG,
    ) -> None:
        super().__init__()
        self.v_rd_c = v_rd_c
        self.d = d
        self.s_r = s_r
        self.a_sw = a_sw
        self.f_ywd_ef = f_ywd_ef
        self.u_1 = u_1
        self.alpha = alpha

    @staticmethod
    def _evaluate(
        v_rd_c: MPA,
        d: MM,
        s_r: MM,
        a_sw: MM2,
        f_ywd_ef: MPA,
        u_1: MM,
        alpha: DEG,
    ) -> MPA:
        """Evaluates the formula, for more information see the __init__ method."""
        raise_if_negative(v_rd_c=v_rd_c, d=d, s_r=s_r, a_sw=a_sw, f_ywd_ef=f_ywd_ef, u_1=u_1, alpha=alpha)

        return 0.75 * v_rd_c + 1.5 * (d / s_r) * a_sw * f_ywd_ef * (1 / (u_1 * d)) * np.sin(np.deg2rad(alpha))

    def latex(self, n: int = 3) -> LatexFormula:
        """Returns LatexFormula object for formula 6.52."""
        _equation: str = (
            r"0.75 \cdot v_{Rd,c} + 1.5 \cdot \frac{ d}{s_r} \cdot A_{sw} \cdot f_{ywd,ef} \cdot \frac{1}{u_{1} \cdot d} \cdot \sin(\alpha)"
        )
        _numeric_equation: str = latex_replace_symbols(
            _equation,
            {
                r"v_{Rd,c}": f"{self.v_rd_c:.{n}f}",
                r"s_r": f"{self.s_r:.{n}f}",
                r"A_{sw}": f"{self.a_sw:.{n}f}",
                r"f_{ywd,ef}": f"{self.f_ywd_ef:.{n}f}",
                r"u_{1}": f"{self.u_1:.{n}f}",
                r"\alpha": f"{self.alpha:.{n}f}",
                r" d": f" {self.d:.{n}f}",
            },
            False,
        )
        return LatexFormula(
            return_symbol=r"v_{Rd,cs}",
            result=f"{self._evaluate(self.v_rd_c, self.d, self.s_r, self.a_sw, self.f_ywd_ef, self.u_1, self.alpha):.{n}f}",
            equation=_equation,
            numeric_equation=_numeric_equation,
            comparison_operator_label="=",
            unit="MPa",
        )


"""Formula 6.52sub1 from EN 1992-1-1:2004: Chapter 6 - Ultimate Limit State."""


class Form6Dot52Sub1EffectiveYieldStrength(Formula):
    r"""Class representing formula 6.52sub1 for the calculation of [$f_{ywd,ef}$]."""

    label = "6.52sub1"
    source_document = EN_1992_1_1_2004

    def __init__(
        self,
        d: MM,
        f_ywd: MPA,
    ) -> None:
        r"""[f_{ywd,ef}] Calculation of [$f_{ywd,ef}$].

        EN 1992-1-1:2004 art.6.4.5(1) - Formula (6.52sub1)

        Parameters
        ----------
        d : MM
            [$d$] Mean effective depth of the slab [$mm$].
        f_ywd : MPA
            [$f_{ywd}$] Design yield strength of the reinforcement [$MPa$].
        """
        super().__init__()
        self.d = d
        self.f_ywd = f_ywd

    @staticmethod
    def _evaluate(
        d: MM,
        f_ywd: MPA,
    ) -> MPA:
        """Evaluates the formula, for more information see the __init__ method."""
        raise_if_negative(d=d, f_ywd=f_ywd)

        return min(250 + 0.25 * d, f_ywd)

    def latex(self, n: int = 3) -> LatexFormula:
        """Returns LatexFormula object for formula 6.52sub1."""
        _equation: str = r"\min\left(250 + 0.25 \cdot d, f_{ywd}\right)"
        _numeric_equation: str = latex_replace_symbols(
            _equation,
            {
                r"f_{ywd}": f"{self.f_ywd:.{n}f}",
                r" d": f" {self.d:.{n}f}",
            },
            False,
        )
        return LatexFormula(
            return_symbol=r"f_{ywd,ef}",
            result=f"{self:.{n}f}",
            equation=_equation,
            numeric_equation=_numeric_equation,
            comparison_operator_label="=",
            unit="MPa",
        )



"""Testing formula 6.12 of EN 1993-1-1:2005."""

import pytest

from blueprints.codes.eurocode.en_1993_1_1_2005.chapter_6_ultimate_limit_state.formula_6_12 import Form6Dot12CheckBendingMoment
from blueprints.validations import LessOrEqualToZeroError, NegativeValueError


class TestForm6Dot12CheckBendingMoment:
    """Validation for formula 6.12 from EN 1993-1-1:2005."""

    def test_evaluation(self) -> None:
        """Tests the evaluation of the result."""
        # Example values
        m_ed = 1000.0
        m_c_rd = 1500.0

        # Object to test
        formula = Form6Dot12CheckBendingMoment(m_ed=m_ed, m_c_rd=m_c_rd)

        # Expected result, manually calculated
        expected_result = True

        assert formula == expected_result

    @pytest.mark.parametrize(
        ("m_ed", "m_c_rd"),
        [
            (1000.0, 0.0),  # m_c_rd is zero
            (-1000.0, 1500.0),  # m_ed is negative
            (1000.0, -1500.0),  # m_c_rd is negative
        ],
    )
    def test_raise_error_when_invalid_values_are_given(self, m_ed: float, m_c_rd: float) -> None:
        """Test invalid values."""
        with pytest.raises((NegativeValueError, LessOrEqualToZeroError)):
            Form6Dot12CheckBendingMoment(m_ed=m_ed, m_c_rd=m_c_rd)

    @pytest.mark.parametrize(
        ("representation", "expected"),
        [
            (
                "complete",
                r"CHECK \to \left( \frac{M_{Ed}}{M_{c,Rd}} \leq 1 \right) \to "
                r"\left( \frac{1000.000}{1500.000} \leq 1 \right) \to OK",
            ),
            ("short", r"CHECK \to OK"),
        ],
    )
    def test_latex(self, representation: str, expected: str) -> None:
        """Test the latex representation of the formula."""
        # Example values
        m_ed = 1000.0
        m_c_rd = 1500.0

        # Object to test
        latex = Form6Dot12CheckBendingMoment(m_ed=m_ed, m_c_rd=m_c_rd).latex()

        actual = {
            "complete": latex.complete,
            "short": latex.short,
        }

        assert expected == actual[representation], f"{representation} representation failed."



"""Testing formula 6.27 of EN 1993-1-1:2005."""

import pytest

from blueprints.codes.eurocode.en_1993_1_1_2005.chapter_6_ultimate_limit_state.formula_6_27 import Form6Dot27VplTRdChannelSection
from blueprints.validations import LessOrEqualToZeroError, NegativeValueError


class TestForm6Dot27VplTRdChannelSection:
    """Validation for formula 6.27 from EN 1993-1-1:2005."""

    def test_evaluation(self) -> None:
        """Tests the evaluation of the result."""
        # Example values
        tau_t_ed = 100.0
        f_y = 355.0
        gamma_m0 = 1.0
        tau_w_ed = 50.0
        v_pl_rd = 200000.0

        # Object to test
        formula = Form6Dot27VplTRdChannelSection(
            tau_t_ed=tau_t_ed,
            f_y=f_y,
            gamma_m0=gamma_m0,
            tau_w_ed=tau_w_ed,
            v_pl_rd=v_pl_rd,
        )

        # Expected result, manually calculated
        manually_calculated_result = 107373.685136  # N

        assert formula == pytest.approx(expected=manually_calculated_result, rel=1e-4)

    @pytest.mark.parametrize(
        ("tau_t_ed", "f_y", "gamma_m0", "tau_w_ed", "v_pl_rd"),
        [
            (-100.0, 355.0, 1.0, 50.0, 200000.0),  # tau_t_ed is negative
            (100.0, 355.0, 1.0, 50.0, -200000.0),  # v_pl_rd is negative
            (100.0, 355.0, 1.0, -50.0, 200000.0),  # tau_w_ed is negative
            (1e9, 355.0, 1.0, 50.0, 200000.0),  # square root becomes negative
        ],
    )
    def test_raise_error_when_invalid_values_are_given(self, tau_t_ed: float, f_y: float, gamma_m0: float, tau_w_ed: float, v_pl_rd: float) -> None:
        """Test invalid values."""
        with pytest.raises((NegativeValueError, LessOrEqualToZeroError)):
            Form6Dot27VplTRdChannelSection(
                tau_t_ed=tau_t_ed,
                f_y=f_y,
                gamma_m0=gamma_m0,
                tau_w_ed=tau_w_ed,
                v_pl_rd=v_pl_rd,
            )

    @pytest.mark.parametrize(
        ("tau_t_ed", "f_y", "gamma_m0", "tau_w_ed", "v_pl_rd"),
        [
            (100.0, 0.0, 1.0, 50.0, 200000.0),  # f_y is zero
            (100.0, -355.0, 1.0, 50.0, 200000.0),  # f_y is negative
            (100.0, 355.0, 0.0, 50.0, 200000.0),  # gamma_m0 is zero
            (100.0, 355.0, -1.0, 50.0, 200000.0),  # gamma_m0 is negative
        ],
    )
    def test_raise_error_when_less_or_equal_to_zero_values_are_given(
        self, tau_t_ed: float, f_y: float, gamma_m0: float, tau_w_ed: float, v_pl_rd: float
    ) -> None:
        """Test invalid values."""
        with pytest.raises((NegativeValueError, LessOrEqualToZeroError)):
            Form6Dot27VplTRdChannelSection(
                tau_t_ed=tau_t_ed,
                f_y=f_y,
                gamma_m0=gamma_m0,
                tau_w_ed=tau_w_ed,
                v_pl_rd=v_pl_rd,
            )

    @pytest.mark.parametrize(
        ("representation", "expected"),
        [
            (
                "complete",
                r"V_{pl,T,Rd} = \left( \sqrt{1 - \frac{\tau_{t,Ed}}{1.25 \cdot \left( f_y / \sqrt{3} \right) / \gamma_{M0}}} - "
                r"\frac{\tau_{w,Ed}}{\left( f_y / \sqrt{3} \right) / \gamma_{M0}} \right) \cdot V_{pl,Rd} = "
                r"\left( \sqrt{1 - \frac{100.000}{1.25 \cdot \left( 355.000 / \sqrt{3} \right) / 1.000}} - "
                r"\frac{50.000}{\left( 355.000 / \sqrt{3} \right) / 1.000} \right) \cdot 200000.000 = 107373.685 \ N",
            ),
            ("short", r"V_{pl,T,Rd} = 107373.685 \ N"),
        ],
    )
    def test_latex(self, representation: str, expected: str) -> None:
        """Test the latex representation of the formula."""
        # Example values
        tau_t_ed = 100.0
        f_y = 355.0
        gamma_m0 = 1.0
        tau_w_ed = 50.0
        v_pl_rd = 200000.0

        # Object to test
        latex = Form6Dot27VplTRdChannelSection(
            tau_t_ed=tau_t_ed,
            f_y=f_y,
            gamma_m0=gamma_m0,
            tau_w_ed=tau_w_ed,
            v_pl_rd=v_pl_rd,
        ).latex()

        actual = {
            "complete": latex.complete,
            "short": latex.short,
        }

        assert expected == actual[representation], f"{representation} representation failed."



"""Formula 5.33 from EN 1992-1-1:2004: Chapter 5 - Structural Analysis."""

from blueprints.codes.eurocode.en_1992_1_1_2004 import EN_1992_1_1_2004
from blueprints.codes.formula import Formula
from blueprints.codes.latex_formula import LatexFormula
from blueprints.type_alias import DIMENSIONLESS, KN, M
from blueprints.validations import raise_if_less_or_equal_to_zero


class Form5Dot33NominalSecondOrderMoment(Formula):
    """Class representing formula 5.33 for the calculation of the nominal 2nd order moment, [$M_{2}$]."""

    label = "5.33"
    source_document = EN_1992_1_1_2004

    def __init__(
        self,
        n_ed: KN,
        curvature: DIMENSIONLESS,
        l_o: M,
        c: DIMENSIONLESS,
    ) -> None:
        r"""[$M_{2}$] Nominal 2nd order moment [$kNm$].

        EN 1992-1-1:2004 art.5.8.8.2 - Formula (5.33)

        Parameters
        ----------
        n_ed : KN
            [$N_{Ed}$] Design value of axial force [$kN$].
        curvature : DIMENSIONLESS
            [$\frac{1}{r}$] Curvature (1/r), see 5.8.8.3 [$1/m$].
        l_o : M
            [$l_{o}$] Effective length, see 5.8.3.2 [$m$].
        c : DIMENSIONLESS
            [$c$] Factor depending on the curvature distribution, see 5.8.8.2 (4). [-].
        """
        super().__init__()
        self.n_ed = n_ed
        self.curvature = curvature
        self.l_o = l_o
        self.c = c

    @staticmethod
    def _evaluate(
        n_ed: KN,
        curvature: float,
        l_o: M,
        c: float,
    ) -> KN:
        """Evaluates the formula, for more information see the __init__ method."""
        raise_if_less_or_equal_to_zero(n_ed=n_ed, curvature=curvature, l_o=l_o, c=c)

        return n_ed * curvature * (l_o**2) / c

    def latex(self, n: int = 3) -> LatexFormula:
        """Returns LatexFormula object for formula 5.33."""
        return LatexFormula(
            return_symbol=r"M_{2}",
            result=f"{self:.{n}f}",
            equation=r"N_{Ed} \cdot \left(\frac{1}{r}\right) \cdot \frac{l_{o}^2}{c}",
            numeric_equation=rf"{self.n_ed:.{n}f} \cdot \left({self.curvature:.{n}f}\right) " rf"\cdot \frac{{{self.l_o:.{n}f}^2}}{{{self.c:.{n}f}}}",
            comparison_operator_label="=",
            unit="kNm",
        )



"""Formula 8.9 from EN 1992-1-1:2004: Chapter 8: Detailing of reinforcement and prestressing tendons."""

from blueprints.codes.eurocode.en_1992_1_1_2004 import EN_1992_1_1_2004
from blueprints.codes.formula import Formula
from blueprints.codes.latex_formula import LatexFormula
from blueprints.type_alias import KN, MM, MM2, MPA
from blueprints.unit_conversion import N_TO_KN
from blueprints.validations import raise_if_less_or_equal_to_zero, raise_if_negative


class Form8Dot9AnchorageCapacityWeldedTransverseBarSmallDiameter(Formula):
    """Class representing the formula 8.9 for the calculation of the anchorage capacity of a welded cross bar for nominal bar diameters smaller
    than 12 mm.
    """

    label = "8.9"
    source_document = EN_1992_1_1_2004

    def __init__(
        self,
        f_wd: KN,
        diameter_t: MM,
        diameter_l: MM,
        a_s: MM2,
        f_cd: MPA,
    ) -> None:
        r"""[$F_{btd}$] Anchorage capacity of a welded cross bar for nominal bar diameters smaller than 12 mm [$kN$].

        EN 1992-1-1:2004 art.8.6(5) - formula 8.9

        Parameters
        ----------
        f_wd : KN
            [$F_{wd}$] Design shear strength of weld (specified as a factor times [$A_{s} \cdot f_{yd}$]; say [$0.5 \cdot A_{s} \cdot f_{yd}$]
            where [$A_{s}$] is the cross-section of the anchored bar and [$f_{yd}$] is its design yield strength)  [$kN$].
        diameter_t : MM
            [$ø_{t}$] Diameter of the transverse bar [$mm$].

            Note: [$ø_{t} \leq 12$] [$mm$].
        diameter_l : MM
            [$ø_{l}$] Diameter of the bar to be anchored [$mm$].

            Note: [$ø_{l} \leq 12$] [$mm$].
        a_s : MM2
            [$A_{s}$] Cross-section of the anchored bar [$mm^{2}$].
        f_cd : MPA
            [$f_{cd}$] Design compressive strength of concrete [$MPa$].
        """
        super().__init__()
        self.f_wd = f_wd
        self.diameter_t = diameter_t
        self.diameter_l = diameter_l
        self.a_s = a_s
        self.f_cd = f_cd

    @staticmethod
    def _evaluate(
        f_wd: KN,
        diameter_t: MM,
        diameter_l: MM,
        a_s: MM2,
        f_cd: MPA,
    ) -> MM:
        """Evaluates the formula, for more information see the __init__ method."""
        raise_if_negative(
            f_wd=f_wd,
            diameter_t=diameter_t,
            a_s=a_s,
            f_cd=f_cd,
        )
        raise_if_less_or_equal_to_zero(diameter_l=diameter_l)
        return min(f_wd, N_TO_KN * 16 * a_s * f_cd * (diameter_t / diameter_l))

    def latex(self, n: int = 2) -> LatexFormula:
        """Returns LatexFormula object for formula 8.9."""
        return LatexFormula(
            return_symbol=r"F_{btd}",
            result=f"{self:.{n}f}",
            equation=r"\min \left( F_{wd}, 16 \cdot A_s \cdot f_{cd} \cdot \frac{Ø_t}{Ø_l} \right)",
            numeric_equation=(
                rf"\min \left( {self.f_wd:.{n}f}, 1000 \cdot 16 \cdot {self.a_s:.{n}f} \cdot {self.f_cd:.{n}f} \cdot "
                rf"\frac{{{self.diameter_t:.{n}f}}}{{{self.diameter_l:.{n}f}}} \right)"
            ),
            comparison_operator_label="=",
        )



"""Testing formula 8.1 of EN 1992-1-1:2004."""

import pytest

from blueprints.codes.eurocode.en_1992_1_1_2004.chapter_8_detailing_of_reinforcement_and_prestressing_tendons.formula_8_1 import (
    Form8Dot1RequiredMinimumMandrelDiameter,
)
from blueprints.validations import NegativeValueError


class TestForm8Dot1RequiredMinimumMandrelDiameter:
    """Validation for formula 8.1 from EN 1992-1-1:2004."""

    def test_evaluation(self) -> None:
        """Test the evaluation of the result."""
        # Example values
        f_bt = 80  # kN
        a_b = 200  # mm
        diameter = 16  # mm
        f_cd = 30  # MPa
        form_8_1 = Form8Dot1RequiredMinimumMandrelDiameter(
            f_bt=f_bt,
            a_b=a_b,
            diameter=diameter,
            f_cd=f_cd,
        )
        # Expected result, manually calculated
        manually_calculated_result = 290 / 3

        assert form_8_1 == pytest.approx(expected=manually_calculated_result, rel=1e-4)

    def test_raise_error_when_negative_f_bt_is_given(self) -> None:
        """Tests if NegativeValueError is raised when f_bt is negative."""
        # Example values
        f_bt = -80  # kN
        a_b = 200  # mm
        diameter = 16  # mm
        f_cd = 30  # MPa

        with pytest.raises(NegativeValueError):
            Form8Dot1RequiredMinimumMandrelDiameter(
                f_bt=f_bt,
                a_b=a_b,
                diameter=diameter,
                f_cd=f_cd,
            )

    def test_raise_error_when_negative_a_b_is_given(self) -> None:
        """Tests if NegativeValueError is raised when a_b is negative."""
        # Example values
        f_bt = 80  # kN
        a_b = -200  # mm
        diameter = 16  # mm
        f_cd = 30  # MPa

        with pytest.raises(NegativeValueError):
            Form8Dot1RequiredMinimumMandrelDiameter(
                f_bt=f_bt,
                a_b=a_b,
                diameter=diameter,
                f_cd=f_cd,
            )

    def test_raise_error_when_negative_diameter_is_given(self) -> None:
        """Tests if NegativeValueError is raised when diameter is negative."""
        # Example values
        f_bt = 80  # kN
        a_b = 200  # mm
        diameter = -16  # mm
        f_cd = 30  # MPa

        with pytest.raises(NegativeValueError):
            Form8Dot1RequiredMinimumMandrelDiameter(
                f_bt=f_bt,
                a_b=a_b,
                diameter=diameter,
                f_cd=f_cd,
            )

    def test_raise_error_when_negative_f_cd_is_given(self) -> None:
        """Tests if NegativeValueError is raised when f_cd is negative."""
        # Example values
        f_bt = 80  # kN
        a_b = 200  # mm
        diameter = 16  # mm
        f_cd = -30  # MPa

        with pytest.raises(NegativeValueError):
            Form8Dot1RequiredMinimumMandrelDiameter(
                f_bt=f_bt,
                a_b=a_b,
                diameter=diameter,
                f_cd=f_cd,
            )

    @pytest.mark.parametrize(
        ("representation", "expected"),
        [
            (
                "complete",
                (
                    r"Ø_{m,min} = \frac{F_{bt} \left( \frac{1}{a_b} + \frac{1}{2 \cdot Ø} \right) }{f_{cd}} "
                    r"= \frac{80.000 \cdot 1000 \cdot \left( \frac{1}{200.000} + \frac{1}{2 \cdot 16.000} \right)}{30.000} = 96.667"
                ),
            ),
            ("short", r"Ø_{m,min} = 96.667"),
        ],
    )
    def test_latex(self, representation: str, expected: str) -> None:
        """Test the latex representation of the formula."""
        # Example values
        f_bt = 80  # kN
        a_b = 200  # mm
        diameter = 16  # mm
        f_cd = 30  # MPa

        # Object to test
        form_8_1_latex = Form8Dot1RequiredMinimumMandrelDiameter(
            f_bt=f_bt,
            a_b=a_b,
            diameter=diameter,
            f_cd=f_cd,
        ).latex()

        actual = {"complete": form_8_1_latex.complete, "short": form_8_1_latex.short}

        assert actual[representation] == expected, f"{representation} representation failed."



"""Tests for geometry operations."""

import math

import pytest
from shapely import LinearRing, Point

from blueprints.geometry.operations import CoordinateSystemOptions, calculate_rotation_angle, rotate_linearring
from blueprints.type_alias import RAD


class TestCalculateRotationAngle:
    """Tests for the calculate_rotation_angle function."""

    @pytest.mark.parametrize(
        ("start_point", "end_point"),
        [
            (Point(0, 0), Point(0, 0)),
            (Point(1, 1), Point(1, 1)),
        ],
    )
    def test_identical_points(self, start_point: Point, end_point: Point) -> None:
        """Test if ValueError is raised when start and end points are the same."""
        with pytest.raises(ValueError):
            calculate_rotation_angle(start_point=start_point, end_point=end_point)

    @pytest.mark.parametrize(
        ("start_point", "end_point", "coordinate_system"),
        [
            (Point(0, 0), Point(1, 1), CoordinateSystemOptions.XZ),
            (Point(0, 0), Point(1, 1), CoordinateSystemOptions.YZ),
        ],
    )
    def test_missing_z_value(self, start_point: Point, end_point: Point, coordinate_system: CoordinateSystemOptions) -> None:
        """Test if ValueError is raised when start or end point do not have z value when rotation
        angle in XZ or YZ plane is requested.
        """
        with pytest.raises(ValueError):
            calculate_rotation_angle(
                start_point=start_point,
                end_point=end_point,
                coordinate_system=coordinate_system,
            )

    @pytest.mark.parametrize(
        ("start_point", "end_point", "coordinate_system", "expected_result"),
        [
            (Point(0, 0), Point(1, 0), CoordinateSystemOptions.XY, 0),
            (Point(0, 0), Point(1, 1), CoordinateSystemOptions.XY, math.pi / 4),
            (Point(0, 0), Point(0, 1), CoordinateSystemOptions.XY, math.pi / 2),
            (Point(0, 0), Point(-1, 1), CoordinateSystemOptions.XY, (3 / 4) * math.pi),
            (Point(0, 0), Point(-1, 0), CoordinateSystemOptions.XY, math.pi),
            (Point(0, 0), Point(-1, -1), CoordinateSystemOptions.XY, (5 / 4) * math.pi),
            (Point(0, 0), Point(0, -1), CoordinateSystemOptions.XY, (3 / 2) * math.pi),
            (Point(0, 0), Point(1, -1), CoordinateSystemOptions.XY, (7 / 4) * math.pi),
            (Point(0, 0, 0), Point(1, 0, 0), CoordinateSystemOptions.XZ, 0),
            (Point(0, 0, 0), Point(1, 0, 1), CoordinateSystemOptions.XZ, math.pi / 4),
            (Point(0, 0, 0), Point(0, 0, 1), CoordinateSystemOptions.XZ, math.pi / 2),
            (Point(0, 0, 0), Point(-1, 0, 1), CoordinateSystemOptions.XZ, (3 / 4) * math.pi),
            (Point(0, 0, 0), Point(-1, 0, 0), CoordinateSystemOptions.XZ, math.pi),
            (Point(0, 0, 0), Point(-1, 0, -1), CoordinateSystemOptions.XZ, (5 / 4) * math.pi),
            (Point(0, 0, 0), Point(0, 0, -1), CoordinateSystemOptions.XZ, (3 / 2) * math.pi),
            (Point(0, 0, 0), Point(1, 0, -1), CoordinateSystemOptions.XZ, (7 / 4) * math.pi),
            (Point(0, 0, 0), Point(0, 1, 0), CoordinateSystemOptions.YZ, 0),
            (Point(0, 0, 0), Point(0, 1, 1), CoordinateSystemOptions.YZ, math.pi / 4),
            (Point(0, 0, 0), Point(0, 0, 1), CoordinateSystemOptions.YZ, math.pi / 2),
            (Point(0, 0, 0), Point(0, -1, 1), CoordinateSystemOptions.YZ, (3 / 4) * math.pi),
            (Point(0, 0, 0), Point(0, -1, 0), CoordinateSystemOptions.YZ, math.pi),
            (Point(0, 0, 0), Point(0, -1, -1), CoordinateSystemOptions.YZ, (5 / 4) * math.pi),
            (Point(0, 0, 0), Point(0, 0, -1), CoordinateSystemOptions.YZ, (3 / 2) * math.pi),
            (Point(0, 0, 0), Point(0, 1, -1), CoordinateSystemOptions.YZ, (7 / 4) * math.pi),
        ],
    )
    def test_keypoints_coordinate_system(
        self,
        start_point: Point,
        end_point: Point,
        coordinate_system: CoordinateSystemOptions,
        expected_result: RAD,
    ) -> None:
        """Test if the rotation angle is calculated correctly for the given keypoints and coordinate system."""
        result = calculate_rotation_angle(start_point=start_point, end_point=end_point, coordinate_system=coordinate_system)
        assert result == expected_result


class TestRotateLinearRing:
    """Tests for the rotate_linearring function."""

    @pytest.mark.parametrize(
        ("linearring", "angle_degrees", "expected_coords"),
        [
            # Test case: No rotation (0 degrees)
            (LinearRing([(1, 0), (0, 1), (-1, 0), (0, -1), (1, 0)]), 0, [(1, 0), (0, 1), (-1, 0), (0, -1), (1, 0)]),
            # Test case: 90 degrees rotation
            (LinearRing([(1, 0), (0, 1), (-1, 0), (0, -1), (1, 0)]), 90, [(0, 1), (-1, 0), (0, -1), (1, 0), (0, 1)]),
            # Test case: 180 degrees rotation
            (LinearRing([(1, 0), (0, 1), (-1, 0), (0, -1), (1, 0)]), 180, [(-1, 0), (0, -1), (1, 0), (0, 1), (-1, 0)]),
            # Test case: 270 degrees rotation
            (LinearRing([(1, 0), (0, 1), (-1, 0), (0, -1), (1, 0)]), 270, [(0, -1), (1, 0), (0, 1), (-1, 0), (0, -1)]),
            # Test case: 360 degrees rotation (full rotation)
            (LinearRing([(1, 0), (0, 1), (-1, 0), (0, -1), (1, 0)]), 360, [(1, 0), (0, 1), (-1, 0), (0, -1), (1, 0)]),
            # Test case: 45 degrees rotation
            (
                LinearRing([(1, 0), (0, 1), (-1, 0), (0, -1), (1, 0)]),
                45,
                [
                    (0.7071067811865476, 0.7071067811865476),
                    (-0.7071067811865476, 0.7071067811865476),
                    (-0.7071067811865476, -0.7071067811865476),
                    (0.7071067811865476, -0.7071067811865476),
                    (0.7071067811865476, 0.7071067811865476),
                ],
            ),
            # Test case: Negative rotation (-90 degrees)
            (LinearRing([(1, 0), (0, 1), (-1, 0), (0, -1), (1, 0)]), -90, [(0, -1), (1, 0), (0, 1), (-1, 0), (0, -1)]),
        ],
    )
    def test_rotation(self, linearring: LinearRing, angle_degrees: float, expected_coords: list) -> None:
        """Test if the LinearRing is rotated correctly."""
        rotated = rotate_linearring(linearring=linearring, angle_degrees=angle_degrees)
        for obtained, expected in zip(rotated.coords, expected_coords):
            assert obtained == pytest.approx(expected, rel=1e-4)



"""Formula 5.14 from EN 1992-1-1:2004: Chapter 5 - Structural Analysis."""

from blueprints.codes.eurocode.en_1992_1_1_2004 import EN_1992_1_1_2004
from blueprints.codes.formula import Formula
from blueprints.codes.latex_formula import LatexFormula
from blueprints.type_alias import DIMENSIONLESS, M
from blueprints.validations import raise_if_less_or_equal_to_zero


class Form5Dot14SlendernessRatio(Formula):
    r"""Class representing formula 5.14 for the calculation of the slenderness ratio, [$\lambda$]."""

    label = "5.14"
    source_document = EN_1992_1_1_2004

    def __init__(
        self,
        l_0: M,
        i: M,
    ) -> None:
        r"""[$\lambda$] Slenderness ratio [$-$].

        EN 1992-1-1:2004 art.5.8.3.2(1) - Formula (5.14)

        Parameters
        ----------
        l_0 : M
            [$l_{0}$] Effective length [$m$].
            Use your own implementation of this value or use :class: `Form5Dot15EffectiveLengthBraced`
            or :class: `Form5Dot15EffectiveLengthUnbraced`.
        i : M
            [$i$] Radius of gyration of the uncracked concrete section [$m$].
        """
        super().__init__()
        self.l_0 = l_0
        self.i = i

    @staticmethod
    def _evaluate(
        l_0: M,
        i: M,
    ) -> DIMENSIONLESS:
        """Evaluates the formula, for more information see the __init__ method."""
        raise_if_less_or_equal_to_zero(
            l_0=l_0,
            i=i,
        )
        return l_0 / i

    def latex(self, n: int = 3) -> LatexFormula:
        """Returns LatexFormula object for formula 5.14."""
        return LatexFormula(
            return_symbol=r"λ",
            result=f"{self:.{n}f}",
            equation=r"\frac{l_0}{i}",
            numeric_equation=rf"\frac{{{self.l_0:.{n}f}}}{{{self.i:.{n}f}}}",
            comparison_operator_label="=",
        )



"""Testing formula 6.11a/b/cN of EN 1992-1-1:2004."""

import pytest

from blueprints.codes.eurocode.en_1992_1_1_2004.chapter_6_ultimate_limit_state.formula_6_11abcn import (
    Form6Dot11abcNCompressionChordCoefficient,
)
from blueprints.validations import LessOrEqualToZeroError, NegativeValueError


class TestForm6Dot11abcNCompressionChordCoefficient:
    """Validation for formula 6.11a/b/cN from EN 1992-1-1:2004."""

    def test_evaluation_below_0_25(self) -> None:
        """Tests the evaluation of the result for sigma_cp/f_cd < 0.25."""
        # Example values
        sigma_cp = 5.0
        f_cd = 30.0

        # Object to test
        formula = Form6Dot11abcNCompressionChordCoefficient(sigma_cp=sigma_cp, f_cd=f_cd)

        # Expected result, manually calculated
        manually_calculated_result = 1 + 1 / 6.0

        assert formula == pytest.approx(expected=manually_calculated_result, rel=1e-4)

    def test_evaluation_between_0_25_and_0_5(self) -> None:
        """Tests the evaluation of the result for 0.25 <= sigma_cp/f_cd <= 0.5."""
        # Example values
        sigma_cp = 10.0
        f_cd = 30.0

        # Object to test
        formula = Form6Dot11abcNCompressionChordCoefficient(sigma_cp=sigma_cp, f_cd=f_cd)

        # Expected result, manually calculated
        manually_calculated_result = 1.25

        assert formula == pytest.approx(expected=manually_calculated_result, rel=1e-4)

    def test_evaluation_above_0_5(self) -> None:
        """Tests the evaluation of the result for sigma_cp/f_cd > 0.5."""
        # Example values
        sigma_cp = 20.0
        f_cd = 30.0

        # Object to test
        formula = Form6Dot11abcNCompressionChordCoefficient(sigma_cp=sigma_cp, f_cd=f_cd)

        # Expected result, manually calculated
        manually_calculated_result = 5 / 6.0

        assert formula == pytest.approx(expected=manually_calculated_result, rel=1e-4)

    @pytest.mark.parametrize(
        ("sigma_cp", "f_cd"),
        [
            (-5.0, 30.0),  # sigma_cp is negative
            (0.0, 30.0),  # sigma_cp is zero
            (5.0, -30.0),  # f_cd is negative
            (5.0, 0.0),  # f_cd is zero
        ],
    )
    def test_raise_error_when_invalid_values_are_given(self, sigma_cp: float, f_cd: float) -> None:
        """Test invalid values."""
        with pytest.raises((NegativeValueError, LessOrEqualToZeroError)):
            Form6Dot11abcNCompressionChordCoefficient(sigma_cp=sigma_cp, f_cd=f_cd)

    @pytest.mark.parametrize(
        ("representation", "expected"),
        [
            (
                "complete",
                r"\alpha_{cw} = \begin{cases} 1 + \frac{\sigma_{cp}}{f_{cd}} & \text{if } 0 \lt \sigma_{cp} \leq 0.25 f_{cd} "
                r"\\ 1.25 & \text{if } 0.25 f_{cd} \lt \sigma_{cp} \leq 0.5 f_{cd} \\ 2.5 \left(1 - \frac{\sigma_{cp}}{f_{cd}}\right) "
                r"& \text{if } \sigma_{cp} \gt 0.5 f_{cd} \end{cases} = "
                r"\begin{cases} 1 + \frac{5.000}{30.000} & \text{if } 0 \lt 5.000 \leq 0.25 \cdot 30.000 \\ 1.250 & \text{if } "
                r"0.25 \cdot 30.000 \lt 5.000 \leq 0.5 \cdot 30.000 \\ 2.5 \left(1 - \frac{5.000}{30.000}\right) & \text{if } "
                r"5.000 \gt 0.5 \cdot 30.000 \end{cases} = 1.167 \ -",
            ),
            ("short", r"\alpha_{cw} = 1.167 \ -"),
        ],
    )
    def test_latex(self, representation: str, expected: str) -> None:
        """Test the latex representation of the formula."""
        # Example values
        sigma_cp = 5.0
        f_cd = 30.0

        # Object to test
        latex = Form6Dot11abcNCompressionChordCoefficient(sigma_cp=sigma_cp, f_cd=f_cd).latex()

        actual = {
            "complete": latex.complete,
            "short": latex.short,
        }

        assert expected == actual[representation], f"{representation} representation failed."



"""Subformula a trough g from 6.18 from EN 1993-1-1:2005: Chapter 6 - Ultimate Limit State."""

import numpy as np

from blueprints.codes.eurocode.en_1993_1_1_2005 import EN_1993_1_1_2005
from blueprints.codes.formula import Formula
from blueprints.codes.latex_formula import LatexFormula, latex_replace_symbols
from blueprints.type_alias import DIMENSIONLESS, MM, MM2
from blueprints.validations import raise_if_less_or_equal_to_zero, raise_if_lists_differ_in_length, raise_if_negative


class Form6Dot18SubARolledIandHSection(Formula):
    r"""Class representing formula 6.18suba for the calculation of shear area for a rolled I and H section."""

    label = "6.18suba"
    source_document = EN_1993_1_1_2005

    def __init__(
        self,
        a: MM2,
        b: MM,
        hw: MM,
        r: MM,
        tf: MM,
        tw: MM,
        eta: DIMENSIONLESS,
    ) -> None:
        r"""[$A_v$] Calculation of the shear area for a rolled I and H section with load parallel to web [$mm^2$].

        EN 1993-1-1:2005 art.6.2.6(3) - Formula (6.18suba)

        Parameters
        ----------
        a : MM2
            [$A$] Cross-sectional area [$mm^2$].
        b : MM
            [$b$] Overall breadth [$mm$].
        hw : MM
            [$h_w$] Depth of the web [$mm$].
        r : MM
            [$r$] Root radius [$mm$].
        tf : MM
            [$t_f$] Flange thickness [$mm$].
        tw : MM
            [$t_w$] Web thickness [$mm$]. If the web thickness is not constant, tw should be taken as the minimum thickness.
        eta : DIMENSIONLESS, optional
            [$\eta$] Dimensionless conversionfactor, see EN 1993-1-5 5.1. Note, $eta$ may be conservatively taken equal to 1.0.
        """
        super().__init__()
        self.a = a
        self.b = b
        self.hw = hw
        self.r = r
        self.tf = tf
        self.tw = tw
        self.eta = eta

    @staticmethod
    def _evaluate(
        a: MM2,
        b: MM,
        hw: MM,
        r: MM,
        tf: MM,
        tw: MM,
        eta: DIMENSIONLESS,
    ) -> MM2:
        """Evaluates the formula, for more information see the __init__ method."""
        raise_if_negative(a=a, b=b, hw=hw, r=r, tf=tf, tw=tw, eta=eta)

        av = a - 2 * b * tf + (tw + 2 * r) * tf
        av_min = eta * hw * tw

        return max(0, av, av_min)

    def latex(self, n: int = 3) -> LatexFormula:
        """Returns LatexFormula object for formula 6.18suba."""
        _equation: str = r"max(A - 2 \cdot b \cdot t_f + (t_w + 2 \cdot r) \cdot t_f; \eta \cdot h_w \cdot t_w)"
        _numeric_equation: str = latex_replace_symbols(
            _equation,
            {
                r"A": f"{self.a:.{n}f}",
                r"b": f"{self.b:.{n}f}",
                r"h_w": f"{self.hw:.{n}f}",
                r"r": f"{self.r:.{n}f}",
                r"t_f": f"{self.tf:.{n}f}",
                r"t_w": f"{self.tw:.{n}f}",
                r"\eta": f"{self.eta:.{n}f}",
            },
            False,
        )
        return LatexFormula(
            return_symbol=r"A_v",
            result=f"{self:.{n}f}",
            equation=_equation,
            numeric_equation=_numeric_equation,
            comparison_operator_label="=",
            unit="mm^2",
        )


class Form6Dot18SubBRolledChannelSection(Formula):
    r"""Class representing formula 6.18subb for the calculation of shear area for a rolled channel section."""

    label = "6.18subb"
    source_document = EN_1993_1_1_2005

    def __init__(
        self,
        a: MM2,
        b: MM,
        tf: MM,
        tw: MM,
        r: MM,
    ) -> None:
        r"""[$A_v$] Calculation of the shear area for a rolled channel section with load parallel to web [$mm^2$].

        EN 1993-1-1:2005 art.6.2.6(3) - Formula (6.18subb)

        Parameters
        ----------
        a : MM2
            [$A$] Cross-sectional area [$mm^2$].
        b : MM
            [$b$] Overall breadth [$mm$].
        tf : MM
            [$t_f$] Flange thickness [$mm$].
        tw : MM
            [$t_w$] Web thickness [$mm$]. If the web thickness is not constant, tw should be taken as the minimum thickness.
        r : MM
            [$r$] Root radius [$mm$].
        """
        super().__init__()
        self.a = a
        self.b = b
        self.tf = tf
        self.tw = tw
        self.r = r

    @staticmethod
    def _evaluate(
        a: MM2,
        b: MM,
        tf: MM,
        tw: MM,
        r: MM,
    ) -> MM2:
        """Evaluates the formula, for more information see the __init__ method."""
        raise_if_negative(a=a, b=b, tf=tf, tw=tw, r=r)

        return max(0, a - 2 * b * tf + (tw + r) * tf)

    def latex(self, n: int = 3) -> LatexFormula:
        """Returns LatexFormula object for formula 6.18subb."""
        _equation: str = r"A - 2 \cdot b \cdot t_f + (t_w + r) \cdot t_f"
        _numeric_equation: str = latex_replace_symbols(
            _equation,
            {
                r"A": f"{self.a:.{n}f}",
                r"b": f"{self.b:.{n}f}",
                r"t_f": f"{self.tf:.{n}f}",
                r"t_w": f"{self.tw:.{n}f}",
                r"r": f"{self.r:.{n}f}",
            },
            False,
        )
        return LatexFormula(
            return_symbol=r"A_v",
            result=f"{self:.{n}f}",
            equation=_equation,
            numeric_equation=_numeric_equation,
            comparison_operator_label="=",
            unit="mm^2",
        )


class Form6Dot18SubCTSectionRolled(Formula):
    r"""Class representing formula 6.18subc for the calculation of shear area for a rolled T-section with load parallel to web."""

    label = "6.18subc"
    source_document = EN_1993_1_1_2005

    def __init__(
        self,
        a: MM2,
        b: MM,
        tf: MM,
        tw: MM,
        r: MM,
    ) -> None:
        r"""[$A_v$] Calculation of the shear area for a T-section with load parallel to web [$mm^2$].

        EN 1993-1-1:2005 art.6.2.6(3) - Formula (6.18subc)

        Parameters
        ----------
        a : MM2
            [$A$] Cross-sectional area [$mm^2$].
        b : MM
            [$b$] Overall breadth [$mm$].
        tf : MM
            [$t_f$] Flange thickness [$mm$].
        tw : MM
            [$t_w$] Web thickness [$mm$]. If the web thickness is not constant, tw should be taken as the minimum thickness.
        r : MM
            [$r$] Root radius [$mm$].
        """
        super().__init__()
        self.a = a
        self.b = b
        self.tf = tf
        self.tw = tw
        self.r = r

    @staticmethod
    def _evaluate(
        a: MM2,
        b: MM,
        tf: MM,
        tw: MM,
        r: MM,
    ) -> MM2:
        """Evaluates the formula, for more information see the __init__ method."""
        raise_if_negative(a=a, b=b, tf=tf, tw=tw, r=r)

        return max(0, a - b * tf + (tw + 2 * r) * tf / 2)

    def latex(self, n: int = 3) -> LatexFormula:
        """Returns LatexFormula object for formula 6.18subc."""
        _equation: str = r"A - b \cdot t_f + (t_w + 2 \cdot r) \cdot \frac{t_f}{2}"
        _numeric_equation: str = latex_replace_symbols(
            _equation,
            {
                r"A": f"{self.a:.{n}f}",
                r"b": f"{self.b:.{n}f}",
                r"t_f": f"{self.tf:.{n}f}",
                r"t_w": f"{self.tw:.{n}f}",
                r" r": f" {self.r:.{n}f}",
            },
            False,
        )
        return LatexFormula(
            return_symbol=r"A_v",
            result=f"{self:.{n}f}",
            equation=_equation,
            numeric_equation=_numeric_equation,
            comparison_operator_label="=",
            unit="mm^2",
        )


class Form6Dot18SubCTSectionWelded(Formula):
    r"""Class representing formula 6.18subc for the calculation of shear area for a welded T-section with load parallel to web."""

    label = "6.18subc"
    source_document = EN_1993_1_1_2005

    def __init__(
        self,
        tf: MM,
        tw: MM,
        h: MM,
    ) -> None:
        r"""[$A_v$] Calculation of the shear area for a T-section with load parallel to web [$mm^2$].

        EN 1993-1-1:2005 art.6.2.6(3) - Formula (6.18subc)

        Parameters
        ----------
        tf : MM
            [$t_f$] Flange thickness [$mm$].
        tw : MM
            [$t_w$] Web thickness [$mm$]. If the web thickness is not constant, tw should be taken as the minimum thickness.
        h : MM
            [$h$] Overall depth [$mm$].
        """
        super().__init__()
        self.tf = tf
        self.tw = tw
        self.h = h

    @staticmethod
    def _evaluate(
        tf: MM,
        tw: MM,
        h: MM,
    ) -> MM2:
        """Evaluates the formula, for more information see the __init__ method."""
        raise_if_negative(tf=tf, tw=tw, h=h)

        return max(0, tw * (h * tf / 2))

    def latex(self, n: int = 3) -> LatexFormula:
        """Returns LatexFormula object for formula 6.18subc."""
        _equation: str = r"t_w \cdot (h \cdot t_f / 2)"
        _numeric_equation: str = latex_replace_symbols(
            _equation,
            {
                r"t_f": f"{self.tf:.{n}f}",
                r"t_w": f"{self.tw:.{n}f}",
                r"h": f"{self.h:.{n}f}",
            },
            False,
        )
        return LatexFormula(
            return_symbol=r"A_v",
            result=f"{self:.{n}f}",
            equation=_equation,
            numeric_equation=_numeric_equation,
            comparison_operator_label="=",
            unit="mm^2",
        )


class Form6Dot18SubDWeldedIHandBoxSection(Formula):
    r"""Class representing formula 6.18subd for the calculation of shear area for welded I, H, and box sections with load parallel to web."""

    label = "6.18subd"
    source_document = EN_1993_1_1_2005

    def __init__(
        self,
        hw_list: list[MM],
        tw_list: list[MM],
        eta: DIMENSIONLESS,
    ) -> None:
        r"""[$A_v$] Calculation of the shear area for welded I, H, and box sections with load parallel to web [$mm^2$].

        EN 1993-1-1:2005 art.6.2.6(3) - Formula (6.18subd)

        Parameters
        ----------
        hw_list : list[MM]
            [$h_w$] List of depths of the web [$mm$].
        tw_list : list[MM]
            [$t_w$] List of web thicknesses [$mm$]. If the web thickness is not constant, tw should be taken as the minimum thickness.
        eta : DIMENSIONLESS
            [$\eta$] See EN 1993-1-5. Note, $eta$ may be conservatively taken equal to 1.0.
        """
        super().__init__()
        self.hw_list = hw_list
        self.tw_list = tw_list
        self.eta = eta

    @staticmethod
    def _evaluate(
        hw_list: list[MM],
        tw_list: list[MM],
        eta: DIMENSIONLESS,
    ) -> MM2:
        """Evaluates the formula, for more information see the __init__ method."""
        raise_if_negative(eta=eta)
        for h, t in zip(hw_list, tw_list):
            raise_if_negative(h=h, t=t)
        raise_if_lists_differ_in_length(hw_list=hw_list, tw_list=tw_list)

        return max(0, eta * sum(h * t for h, t in zip(hw_list, tw_list)))

    def latex(self, n: int = 3) -> LatexFormula:
        """Returns LatexFormula object for formula 6.18subd."""
        _equation: str = r"\eta \cdot \sum (h_{w} \cdot t_{w})"
        _numeric_equation: str = rf"{self.eta:.{n}f} \cdot (" + rf"{self.hw_list[0]:.{n}f} \cdot {self.tw_list[0]:.{n}f}"
        for i in range(1, len(self.hw_list)):
            _numeric_equation += rf" + {self.hw_list[i]:.{n}f} \cdot {self.tw_list[i]:.{n}f}"
        _numeric_equation += ")"
        return LatexFormula(
            return_symbol=r"A_v",
            result=f"{self:.{n}f}",
            equation=_equation,
            numeric_equation=_numeric_equation,
            comparison_operator_label="=",
            unit="mm^2",
        )


class Form6Dot18SubEWeldedIHandBoxSection(Formula):
    r"""Class representing formula 6.18sube for the calculation of shear area for welded I, H, channel, and box sections with
    load parallel to flanges.
    """

    label = "6.18sube"
    source_document = EN_1993_1_1_2005

    def __init__(
        self,
        a: MM2,
        hw_list: list[MM],
        tw_list: list[MM],
    ) -> None:
        r"""[$A_v$] Calculation of the shear area for welded I, H, channel, and box sections with load parallel to flanges [$mm^2$].

        EN 1993-1-1:2005 art.6.2.6(3) - Formula (6.18sube)

        Parameters
        ----------
        a : MM2
            [$A$] Cross-sectional area [$mm^2$].
        hw_list : list[MM]
            [$h_w$] List of depths of the web [$mm$].
        tw_list : list[MM]
            [$t_w$] List of web thicknesses [$mm$]. If the web thickness is not constant, tw should be taken as the minimum thickness.
        """
        super().__init__()
        self.a = a
        self.hw_list = hw_list
        self.tw_list = tw_list

    @staticmethod
    def _evaluate(
        a: MM2,
        hw_list: list[MM],
        tw_list: list[MM],
    ) -> MM2:
        """Evaluates the formula, for more information see the __init__ method."""
        raise_if_negative(a=a)
        for h, t in zip(hw_list, tw_list):
            raise_if_negative(h=h, t=t)
        raise_if_lists_differ_in_length(hw=hw_list, tw_list=tw_list)

        return max(0, a - sum(h * t for h, t in zip(hw_list, tw_list)))

    def latex(self, n: int = 3) -> LatexFormula:
        """Returns LatexFormula object for formula 6.18sube."""
        _equation: str = r"A - \sum (h_{w} \cdot t_{w})"
        _numeric_equation: str = rf"{self.a:.{n}f} - (" + rf"{self.hw_list[0]:.{n}f} \cdot {self.tw_list[0]:.{n}f}"
        for i in range(1, len(self.hw_list)):
            _numeric_equation += rf" + {self.hw_list[i]:.{n}f} \cdot {self.tw_list[i]:.{n}f}"
        _numeric_equation += ")"
        return LatexFormula(
            return_symbol=r"A_v",
            result=f"{self:.{n}f}",
            equation=_equation,
            numeric_equation=_numeric_equation,
            comparison_operator_label="=",
            unit="mm^2",
        )


class Form6Dot18SubF1RolledRectangularHollowSectionDepth(Formula):
    r"""Class representing formula 6.18subf1 for the calculation of shear area for rolled rectangular hollow sections of uniform thickness with
    load parallel to depth.
    """

    label = "6.18subf1"
    source_document = EN_1993_1_1_2005

    def __init__(
        self,
        a: MM2,
        b: MM,
        h: MM,
    ) -> None:
        r"""[$A_v$] Calculation of the shear area for rolled rectangular hollow sections of uniform thickness with load parallel to depth [$mm^2$].

        EN 1993-1-1:2005 art.6.2.6(3) - Formula (6.18subf1)

        Parameters
        ----------
        a : MM2
            [$A$] Cross-sectional area [$mm^2$].
        b : MM
            [$b$] Overall breadth [$mm$].
        h : MM
            [$h$] Overall depth [$mm$].
        """
        super().__init__()
        self.a = a
        self.b = b
        self.h = h

    @staticmethod
    def _evaluate(
        a: MM2,
        b: MM,
        h: MM,
    ) -> MM2:
        """Evaluates the formula, for more information see the __init__ method."""
        raise_if_negative(a=a, b=b, h=h)
        denominator = b + h
        raise_if_less_or_equal_to_zero(denominator=denominator)
        return max(0, a * h / (b + h))

    def latex(self, n: int = 3) -> LatexFormula:
        """Returns LatexFormula object for formula 6.18subf1."""
        _equation: str = r"\frac{A \cdot h}{b + h}"
        _numeric_equation: str = latex_replace_symbols(
            _equation,
            {
                r"A": f"{self.a:.{n}f}",
                r"b": f"{self.b:.{n}f}",
                r"h": f"{self.h:.{n}f}",
            },
            False,
        )
        return LatexFormula(
            return_symbol=r"A_v",
            result=f"{self:.{n}f}",
            equation=_equation,
            numeric_equation=_numeric_equation,
            comparison_operator_label="=",
            unit="mm^2",
        )


class Form6Dot18SubF2RolledRectangularHollowSectionWidth(Formula):
    r"""Class representing formula 6.18subf2 for the calculation of shear area for rolled rectangular hollow sections of uniform thickness with
    load parallel to width.
    """

    label = "6.18subf2"
    source_document = EN_1993_1_1_2005

    def __init__(
        self,
        a: MM2,
        b: MM,
        h: MM,
    ) -> None:
        r"""[$A_v$] Calculation of the shear area for rolled rectangular hollow sections of uniform thickness with load parallel to width [$mm^2$].

        EN 1993-1-1:2005 art.6.2.6(3) - Formula (6.18subf2)

        Parameters
        ----------
        a : MM2
            [$A$] Cross-sectional area [$mm^2$].
        b : MM
            [$b$] Overall breadth [$mm$].
        h : MM
            [$h$] Overall depth [$mm$].
        """
        super().__init__()
        self.a = a
        self.b = b
        self.h = h

    @staticmethod
    def _evaluate(
        a: MM2,
        b: MM,
        h: MM,
    ) -> MM2:
        """Evaluates the formula, for more information see the __init__ method."""
        raise_if_negative(a=a, b=b, h=h)
        denominator = b + h
        raise_if_less_or_equal_to_zero(denominator=denominator)
        return max(0, a * b / (b + h))

    def latex(self, n: int = 3) -> LatexFormula:
        """Returns LatexFormula object for formula 6.18subf2."""
        _equation: str = r"\frac{A \cdot b}{b + h}"
        _numeric_equation: str = latex_replace_symbols(
            _equation,
            {
                r"A": f"{self.a:.{n}f}",
                r"b": f"{self.b:.{n}f}",
                r"h": f"{self.h:.{n}f}",
            },
            False,
        )
        return LatexFormula(
            return_symbol=r"A_v",
            result=f"{self:.{n}f}",
            equation=_equation,
            numeric_equation=_numeric_equation,
            comparison_operator_label="=",
            unit="mm^2",
        )


class Form6Dot18SubGCircularHollowSection(Formula):
    r"""Class representing formula 6.18subg for the calculation of shear area for circular hollow sections and tubes of uniform thickness."""

    label = "6.18subg"
    source_document = EN_1993_1_1_2005

    def __init__(
        self,
        a: MM2,
    ) -> None:
        r"""[$A_v$] Calculation of the shear area for circular hollow sections and tubes of uniform thickness [$mm^2$].

        EN 1993-1-1:2005 art.6.2.6(3) - Formula (6.18subg)

        Parameters
        ----------
        a : MM2
            [$A$] Cross-sectional area [$mm^2$].
        """
        super().__init__()
        self.a = a

    @staticmethod
    def _evaluate(
        a: MM2,
    ) -> MM2:
        """Evaluates the formula, for more information see the __init__ method."""
        raise_if_negative(a=a)

        return max(0, 2 * a / np.pi)

    def latex(self, n: int = 3) -> LatexFormula:
        """Returns LatexFormula object for formula 6.18subg."""
        _equation: str = r"\frac{2 \cdot A}{\pi}"
        _numeric_equation: str = latex_replace_symbols(
            _equation,
            {
                r"A": f"{self.a:.{n}f}",
            },
            False,
        )
        return LatexFormula(
            return_symbol=r"A_v",
            result=f"{self:.{n}f}",
            equation=_equation,
            numeric_equation=_numeric_equation,
            comparison_operator_label="=",
            unit="mm^2",
        )



"""Formula 6.42 from EN 1992-1-1:2004: Chapter 6 - Ultimate Limit State."""

import numpy as np

from blueprints.codes.eurocode.en_1992_1_1_2004 import EN_1992_1_1_2004
from blueprints.codes.formula import Formula
from blueprints.codes.latex_formula import LatexFormula, latex_replace_symbols
from blueprints.type_alias import DIMENSIONLESS, MM
from blueprints.validations import raise_if_less_or_equal_to_zero, raise_if_negative


class Form6Dot42BetaCircular(Formula):
    r"""Class representing formula 6.42 for the calculation of [$\beta$]."""

    label = "6.42"
    source_document = EN_1992_1_1_2004

    def __init__(
        self,
        d: MM,
        diameter: MM,
        e: MM,
    ) -> None:
        r"""[$\beta$] Calculation of [$\beta$].

        EN 1992-1-1:2004 art.6.4.3(3) - Formula (6.42)

        Parameters
        ----------
        d : MM
            [$d$] Effective depth of the slab [$mm$].
        diameter : MM
            [$D$] Diameter of the circular column [$mm$].
        e : MM
            [$e$] Distance from the axis about which the moment [$M_{Ed}$] acts [$mm$].
        """
        super().__init__()
        self.d = d
        self.diameter = diameter
        self.e = e

    @staticmethod
    def _evaluate(
        d: MM,
        diameter: MM,
        e: MM,
    ) -> DIMENSIONLESS:
        """Evaluates the formula, for more information see the __init__ method."""
        raise_if_negative(d=d, diameter=diameter, e=e)
        denominator = diameter + 4 * d
        raise_if_less_or_equal_to_zero(denominator=denominator)

        return 1 + 0.6 * np.pi * e / (diameter + 4 * d)

    def latex(self, n: int = 3) -> LatexFormula:
        """Returns LatexFormula object for formula 6.42."""
        _equation: str = r"1 + 0.6 \cdot \pi \cdot \frac{e}{D + 4 \cdot d}"
        _numeric_equation: str = latex_replace_symbols(
            _equation,
            {
                r" d": f" {self.d:.{n}f}",
                r"D": f"{self.diameter:.{n}f}",
                r"e": f"{self.e:.{n}f}",
            },
            True,
        )
        return LatexFormula(
            return_symbol=r"\beta",
            result=f"{self:.{n}f}",
            equation=_equation,
            numeric_equation=_numeric_equation,
            comparison_operator_label="=",
            unit="-",
        )



"""Formula 6.77 from EN 1992-1-1:2004: Chapter 6 - Ultimate limit state."""

from blueprints.codes.eurocode.en_1992_1_1_2004 import EN_1992_1_1_2004
from blueprints.codes.formula import Formula
from blueprints.codes.latex_formula import LatexFormula, latex_replace_symbols
from blueprints.type_alias import DIMENSIONLESS, MPA
from blueprints.validations import raise_if_less_or_equal_to_zero, raise_if_negative


class Form6Dot77FatigueVerification(Formula):
    r"""Class representing formula 6.77 for the fatigue verification of concrete."""

    label = "6.77"
    source_document = EN_1992_1_1_2004

    def __init__(
        self,
        sigma_c_max: MPA,
        sigma_c_min: MPA,
        f_cd_fat: MPA,
        f_ck: MPA,
    ) -> None:
        r"""[$\sigma_{Rd,max}$] Fatigue verification for concrete [$-$].

        EN 1992-1-1:2004 art.6.8.7(2) - Formula (6.77)

        Parameters
        ----------
        sigma_c_max : MPA
            [$\sigma_{c,max}$] Maximum compressive stress at a fibre under the frequent load combination [$MPa$].
        sigma_c_min : MPA
            [$\sigma_{c,min}$] Minimum compressive stress at the same fibre where the maximum occurs [$MPa$].
        f_cd_fat : MPA
            [$f_{cd,fat}$] Design compressive strength of concrete under fatigue [$MPa$].
        f_ck : MPA
            [$f_{ck}$] Characteristic compressive strength of concrete [$MPa$].
        """
        super().__init__()
        self.sigma_c_max = sigma_c_max
        self.sigma_c_min = sigma_c_min
        self.f_cd_fat = f_cd_fat
        self.f_ck = f_ck

    @staticmethod
    def _evaluate(
        sigma_c_max: MPA,
        sigma_c_min: MPA,
        f_cd_fat: MPA,
        f_ck: MPA,
    ) -> DIMENSIONLESS:
        """Evaluates the formula, for more information see the __init__ method."""
        raise_if_negative(sigma_c_max=sigma_c_max, sigma_c_min=sigma_c_min, f_ck=f_ck)
        raise_if_less_or_equal_to_zero(f_cd_fat=f_cd_fat)

        return sigma_c_max / f_cd_fat <= min(0.5 + 0.45 * sigma_c_min / f_cd_fat, 0.9 if f_ck <= 50 else 0.8)

    def latex(self, n: int = 3) -> LatexFormula:
        """Returns LatexFormula object for formula 6.77."""
        _equation: str = (
            r"\frac{\sigma_{c,max}}{f_{cd,fat}} \leq \min\left(0.5 + 0.45 \cdot \frac{\sigma_{c,min}}{f_{cd,fat}}, \begin{cases} 0.9 & "
            r"\text{if } f_{ck} \leq 50 \\ 0.8 & \text{if } f_{ck} > 50 \end{cases}\right)"
        )
        _numeric_equation: str = latex_replace_symbols(
            _equation,
            {
                r"\sigma_{c,max}": f"{self.sigma_c_max:.{n}f}",
                r"\sigma_{c,min}": f"{self.sigma_c_min:.{n}f}",
                "f_{cd,fat}": f"{self.f_cd_fat:.{n}f}",
                "f_{ck}": f"{self.f_ck:.{n}f}",
            },
            False,
        )
        return LatexFormula(
            return_symbol=r"CHECK",
            result="OK" if self.__bool__() else "\\text{Not OK}",
            equation=_equation,
            numeric_equation=_numeric_equation,
            comparison_operator_label="\\to",
            unit="",
        )



"""Testing formula 6.47 of EN 1992-1-1:2004."""

import pytest

from blueprints.codes.eurocode.en_1992_1_1_2004.chapter_6_ultimate_limit_state.formula_6_47 import (
    Form6Dot47PunchingShearResistance,
)
from blueprints.validations import NegativeValueError


class TestForm6Dot47PunchingShearResistance:
    """Validation for formula 6.47 from EN 1992-1-1:2004."""

    def test_evaluation(self) -> None:
        """Tests the evaluation of the result."""
        # Example values
        c_rd_c = 0.18
        k = 1.5
        rho_l = 0.02
        f_ck = 30.0
        k_1 = 0.1
        sigma_cp = 1.0
        v_min = 0.035

        # Object to test
        formula = Form6Dot47PunchingShearResistance(c_rd_c=c_rd_c, k=k, rho_l=rho_l, f_ck=f_ck, k_1=k_1, sigma_cp=sigma_cp, v_min=v_min)

        # Expected result, manually calculated
        manually_calculated_result = 1.15701426312  # MPa

        assert formula == pytest.approx(expected=manually_calculated_result, rel=1e-4)

    @pytest.mark.parametrize(
        ("c_rd_c", "k", "rho_l", "f_ck", "k_1", "sigma_cp", "v_min"),
        [
            (-0.18, 1.5, 0.02, 30.0, 0.1, 1.0, 0.035),  # c_rd_c is negative
            (0.18, -1.5, 0.02, 30.0, 0.1, 1.0, 0.035),  # k is negative
            (0.18, 1.5, -0.02, 30.0, 0.1, 1.0, 0.035),  # rho_l is negative
            (0.18, 1.5, 0.02, -30.0, 0.1, 1.0, 0.035),  # f_ck is negative
            (0.18, 1.5, 0.02, 30.0, -0.1, 1.0, 0.035),  # k_1 is negative
            (0.18, 1.5, 0.02, 30.0, 0.1, -1.0, 0.035),  # sigma_cp is negative
            (0.18, 1.5, 0.02, 30.0, 0.1, 1.0, -0.035),  # v_min is negative
        ],
    )
    def test_raise_error_when_invalid_values_are_given(
        self, c_rd_c: float, k: float, rho_l: float, f_ck: float, k_1: float, sigma_cp: float, v_min: float
    ) -> None:
        """Test invalid values."""
        with pytest.raises(NegativeValueError):
            Form6Dot47PunchingShearResistance(c_rd_c=c_rd_c, k=k, rho_l=rho_l, f_ck=f_ck, k_1=k_1, sigma_cp=sigma_cp, v_min=v_min)

    @pytest.mark.parametrize(
        ("representation", "expected"),
        [
            (
                "complete",
                r"v_{Rd,c} = \max \left( C_{Rd,c} \cdot k \cdot (100 \cdot \rho_l \cdot f_{ck})^{1/3} + "
                r"k_1 \cdot \sigma_{cp}, v_{min} + k_1 \cdot \sigma_{cp} \right) = "
                r"\max \left( 0.180 \cdot 1.500 \cdot (100 \cdot 0.020 \cdot 30.000)^{1/3} + "
                r"0.100 \cdot 1.000, 0.035 + 0.100 \cdot 1.000 \right) = 1.157 \ MPa",
            ),
            ("short", r"v_{Rd,c} = 1.157 \ MPa"),
        ],
    )
    def test_latex(self, representation: str, expected: str) -> None:
        """Test the latex representation of the formula."""
        # Example values
        c_rd_c = 0.18
        k = 1.5
        rho_l = 0.02
        f_ck = 30.0
        k_1 = 0.1
        sigma_cp = 1.0
        v_min = 0.035

        # Object to test
        latex = Form6Dot47PunchingShearResistance(c_rd_c=c_rd_c, k=k, rho_l=rho_l, f_ck=f_ck, k_1=k_1, sigma_cp=sigma_cp, v_min=v_min).latex()

        actual = {
            "complete": latex.complete,
            "short": latex.short,
        }

        assert expected == actual[representation], f"{representation} representation failed."



"""Cross-section base class."""

from abc import ABC, abstractmethod

from sectionproperties.analysis import Section
from sectionproperties.post.post import SectionProperties
from sectionproperties.pre import Geometry
from shapely import Point, Polygon

from blueprints.type_alias import MM, MM2


class CrossSection(ABC):
    """Base class for cross-section shapes."""

    @property
    @abstractmethod
    def name(self) -> str:
        """Name of the cross-section."""

    @property
    @abstractmethod
    def polygon(self) -> Polygon:
        """Shapely Polygon representing the cross-section."""

    @property
    def area(self) -> MM2:
        """Area of the cross-section [mm²].

        When using circular cross-sections, the area is an approximation of the area of the polygon.
        The area is calculated using the `area` property of the Shapely Polygon.

        In case you need an exact answer then you need to override this method in the derived class.
        """
        return self.polygon.area

    @property
    def perimeter(self) -> MM:
        """Perimeter of the cross-section [mm]."""
        return self.polygon.length

    @property
    def centroid(self) -> Point:
        """Centroid of the cross-section [mm]."""
        return self.polygon.centroid

    def geometry(self, mesh_size: MM | None = None) -> Geometry:
        """Geometry of the cross-section.

        Properties
        ----------
        mesh_size : MM
            Maximum mesh element area to be used within
            the Geometry-object finite-element mesh. If not provided, a default value will be used.
        """
        if mesh_size is None:
            mesh_size = 2.0

        geom = Geometry(geom=self.polygon)
        geom.create_mesh(mesh_sizes=mesh_size)
        return geom

    def section(self) -> Section:
        """Section object representing the cross-section."""
        return Section(geometry=self.geometry())

    def section_properties(
        self,
        geometric: bool = True,
        plastic: bool = True,
        warping: bool = True,
    ) -> SectionProperties:
        """Calculate and return the section properties of the cross-section.

        Parameters
        ----------
        geometric : bool
            Whether to calculate geometric properties.
        plastic: bool
            Whether to calculate plastic properties.
        warping: bool
            Whether to calculate warping properties.
        """
        section = self.section()

        if any([geometric, plastic, warping]):
            section.calculate_geometric_properties()
        if warping:
            section.calculate_warping_properties()
        if plastic:
            section.calculate_plastic_properties()

        return section.section_props



"""Formula 7.12 from EN 1992-1-1:2004: Chapter 7 - Serviceability Limit State."""

from blueprints.codes.eurocode.en_1992_1_1_2004 import EN_1992_1_1_2004
from blueprints.codes.formula import Formula
from blueprints.codes.latex_formula import LatexFormula, latex_replace_symbols
from blueprints.type_alias import DIMENSIONLESS, MM
from blueprints.validations import raise_if_negative


class Form7Dot12EquivalentDiameter(Formula):
    r"""Class representing formula 7.12 for the calculation of [$⌀_{eq}$]."""

    label = "7.12"
    source_document = EN_1992_1_1_2004

    def __init__(
        self,
        n_1: DIMENSIONLESS,
        diam_1: MM,
        n_2: DIMENSIONLESS,
        diam_2: MM,
    ) -> None:
        r"""[$⌀_{eq}$] Calculation of the equivalent diameter [$mm$].

        EN 1992-1-1:2004 art.7.3.4(3) - Formula (7.12)

        Parameters
        ----------
        n_1 : DIMENSIONLESS
            [$n_1$] Number of bars with diameter $⌀_1$.
        diam_1 : MM
            [$⌀_1$] Diameter of the first set of bars [$mm$].
        n_2 : DIMENSIONLESS
            [$n_2$] Number of bars with diameter $⌀_2$.
        diam_2 : MM
            [$⌀_2$] Diameter of the second set of bars [$mm$].
        """
        super().__init__()
        self.n_1 = n_1
        self.diam_1 = diam_1
        self.n_2 = n_2
        self.diam_2 = diam_2

    @staticmethod
    def _evaluate(
        n_1: DIMENSIONLESS,
        diam_1: MM,
        n_2: DIMENSIONLESS,
        diam_2: MM,
    ) -> MM:
        """Evaluates the formula, for more information see the __init__ method."""
        raise_if_negative(n_1=n_1, diam_1=diam_1, n_2=n_2, diam_2=diam_2)

        return (n_1 * diam_1**2 + n_2 * diam_2**2) / (n_1 * diam_1 + n_2 * diam_2)

    def latex(self, n: int = 3) -> LatexFormula:
        """Returns LatexFormula object for formula 7.12."""
        _equation: str = r"\frac{n_1 \cdot ⌀_1^2 + n_2 \cdot ⌀_2^2}{n_1 \cdot ⌀_1 + n_2 \cdot ⌀_2}"
        _numeric_equation: str = latex_replace_symbols(
            _equation,
            {
                r"n_1": f"{self.n_1:.{n}f}",
                r"⌀_1": f"{self.diam_1:.{n}f}",
                r"n_2": f"{self.n_2:.{n}f}",
                r"⌀_2": f"{self.diam_2:.{n}f}",
            },
            False,
        )
        return LatexFormula(
            return_symbol=r"⌀_{eq}",
            result=f"{self:.{n}f}",
            equation=_equation,
            numeric_equation=_numeric_equation,
            comparison_operator_label="=",
            unit="mm",
        )



"""Formula 6.32 from EN 1992-1-1:2004: Chapter 6 - Ultimate limit state."""

from blueprints.codes.eurocode.en_1992_1_1_2004 import EN_1992_1_1_2004
from blueprints.codes.formula import Formula
from blueprints.codes.latex_formula import LatexFormula
from blueprints.type_alias import MM
from blueprints.validations import raise_if_negative


class Form6Dot32EffectiveDepthSlab(Formula):
    r"""Class representing formula 6.32 for the calculation of the effective depth of the slab, [$d_{eff}$]."""

    label = "6.32"
    source_document = EN_1992_1_1_2004

    def __init__(
        self,
        d_y: MM,
        d_z: MM,
    ) -> None:
        r"""[$d_{eff}$] Effective depth of the slab [$mm$].

        EN 1992-1-1:2004 art.6.4.2(1) - Formula (6.32)

        Parameters
        ----------
        d_y : MM
            [$d_{y}$] Effective depth of the reinforcement in the y-direction [$mm$].
        d_z : MM
            [$d_{z}$] Effective depth of the reinforcement in the z-direction [$mm$].
        """
        super().__init__()
        self.d_y = d_y
        self.d_z = d_z

    @staticmethod
    def _evaluate(
        d_y: MM,
        d_z: MM,
    ) -> MM:
        """Evaluates the formula, for more information see the __init__ method."""
        raise_if_negative(d_y=d_y, d_z=d_z)
        return (d_y + d_z) / 2

    def latex(self, n: int = 3) -> LatexFormula:
        """Returns LatexFormula object for formula 6.32."""
        return LatexFormula(
            return_symbol=r"d_{eff}",
            result=f"{self:.{n}f}",
            equation=r"\frac{d_{y} + d_{z}}{2}",
            numeric_equation=rf"\frac{{{self.d_y:.{n}f} + {self.d_z:.{n}f}}}{{2}}",
            comparison_operator_label="=",
            unit="mm",
        )



"""Formula 6.33 from EN 1993-1-1:2005: Chapter 6 - Ultimate Limit State."""

from blueprints.codes.eurocode.en_1993_1_1_2005 import EN_1993_1_1_2005
from blueprints.codes.formula import Formula
from blueprints.codes.latex_formula import LatexFormula, latex_replace_symbols
from blueprints.type_alias import N
from blueprints.validations import raise_if_negative


class Form6Dot33CheckAxialForceY(Formula):
    r"""Class representing formula 6.33 for checking axial force about the y-y axis."""

    label = "6.33"
    source_document = EN_1993_1_1_2005

    def __init__(
        self,
        n_ed: N,
        n_pl_rd: N,
    ) -> None:
        r"""For doubly symmetrical I- and H-sections or other flanges sections,
        allowance need not be made for the effect of the axial force on the
        plastic resistance moment about the y-y axis when 6.33 and 6.34 are satisfied.

        EN 1993-1-1:2005 art.6.2.9(4) - Formula (6.33)

        Parameters
        ----------
        n_ed : N
            [$N_{Ed}$] Design axial force [$N$].
        n_pl_rd : N
            [$N_{pl,Rd}$] Plastic resistance normal force [$N$].
        """
        super().__init__()
        self.n_ed = n_ed
        self.n_pl_rd = n_pl_rd

    @staticmethod
    def _evaluate(
        n_ed: N,
        n_pl_rd: N,
    ) -> bool:
        """Evaluates the formula, for more information see the __init__ method."""
        raise_if_negative(n_pl_rd=n_pl_rd)

        return n_ed <= 0.25 * n_pl_rd

    def latex(self, n: int = 3) -> LatexFormula:
        """Returns LatexFormula object for formula 6.33."""
        _equation: str = r"N_{Ed} \leq 0.25 \cdot N_{pl,Rd}"
        _numeric_equation: str = latex_replace_symbols(
            _equation,
            {
                r"N_{Ed}": f"{self.n_ed:.{n}f}",
                r"N_{pl,Rd}": f"{self.n_pl_rd:.{n}f}",
            },
            False,
        )
        _numeric_equation_with_units: str = latex_replace_symbols(
            _equation,
            {
                r"N_{Ed}": rf"{self.n_ed:.{n}f} \ N",
                r"N_{pl,Rd}": rf"{self.n_pl_rd:.{n}f} \ N",
            },
            True,
        )
        return LatexFormula(
            return_symbol="CHECK",
            result="OK" if self.__bool__() else "\\text{Not OK}",
            equation=_equation,
            numeric_equation=_numeric_equation,
            numeric_equation_with_units=_numeric_equation_with_units,
            comparison_operator_label="\\to",
            unit="",
        )



"""Testing formula 6.63 of EN 1992-1-1:2004."""

import pytest

from blueprints.codes.eurocode.en_1992_1_1_2004.chapter_6_ultimate_limit_state.formula_6_63 import Form6Dot63ConcentratedResistanceForce
from blueprints.validations import LessOrEqualToZeroError, NegativeValueError


class TestForm6Dot63ConcentratedResistanceForce:
    """Validation for formula 6.63 from EN 1992-1-1:2004."""

    def test_evaluation(self) -> None:
        """Tests the evaluation of the result."""
        # Example values
        a_c0 = 300.0
        a_c1 = 400.0
        f_cd = 30.0

        # Object to test
        formula = Form6Dot63ConcentratedResistanceForce(a_c0=a_c0, a_c1=a_c1, f_cd=f_cd)

        # Expected result, manually calculated
        manually_calculated_result = 10392.304845413264  # N

        assert formula == pytest.approx(expected=manually_calculated_result, rel=1e-4)

    @pytest.mark.parametrize(
        ("a_c0", "a_c1", "f_cd"),
        [
            (-300.0, 400.0, 30.0),  # a_c0 is negative
            (300.0, -400.0, 30.0),  # a_c1 is negative
            (300.0, 400.0, -30.0),  # f_cd is negative
            (0.0, 400.0, 30.0),  # a_c0 is zero
        ],
    )
    def test_raise_error_when_invalid_values_are_given(self, a_c0: float, a_c1: float, f_cd: float) -> None:
        """Test invalid values."""
        with pytest.raises((NegativeValueError, LessOrEqualToZeroError)):
            Form6Dot63ConcentratedResistanceForce(a_c0=a_c0, a_c1=a_c1, f_cd=f_cd)

    @pytest.mark.parametrize(
        ("representation", "expected"),
        [
            (
                "complete",
                r"F_{Rdu} = \min \left( A_{c0} \cdot f_{cd} \cdot \sqrt{\frac{A_{c1}}{A_{c0}}}, 3 \cdot f_{cd} \cdot A_{c0} \right) = "
                r"\min \left( 300.000 \cdot 30.000 \cdot \sqrt{\frac{400.000}{300.000}}, 3 \cdot 30.000 \cdot 300.000 \right) = 10392.305 \ N",
            ),
            ("short", r"F_{Rdu} = 10392.305 \ N"),
        ],
    )
    def test_latex(self, representation: str, expected: str) -> None:
        """Test the latex representation of the formula."""
        # Example values
        a_c0 = 300.0
        a_c1 = 400.0
        f_cd = 30.0

        # Object to test
        latex = Form6Dot63ConcentratedResistanceForce(a_c0=a_c0, a_c1=a_c1, f_cd=f_cd).latex()

        actual = {
            "complete": latex.complete,
            "short": latex.short,
        }

        assert expected == actual[representation], f"{representation} representation failed."



"""Testing formula 5.37 of EN 1992-1-1:2004."""

import pytest

from blueprints.codes.eurocode.en_1992_1_1_2004.chapter_5_structural_analysis.formula_5_37 import Form5Dot37CreepFactor
from blueprints.validations import LessOrEqualToZeroError, NegativeValueError


class TestForm5Dot37CreepFactor:
    """Validation for formula 5.37 from EN 1992-1-1:2004."""

    def test_evaluation(self) -> None:
        """Tests the evaluation of the result."""
        # Example values
        f_ck = 30.0
        lambda_ = 2.0
        phi_ef = 1.5

        # Object to test
        formula = Form5Dot37CreepFactor(f_ck=f_ck, lambda_=lambda_, phi_ef=phi_ef)

        # Expected result, manually calculated
        manually_calculated_result = 1.730

        assert formula == pytest.approx(expected=manually_calculated_result, rel=1e-4)

    @pytest.mark.parametrize(
        ("f_ck", "lambda_", "phi_ef"),
        [
            (-30.0, 2.0, 1.5),  # f_ck is negative
            (30.0, -2.0, 1.5),  # lambda_ is negative
            (30.0, 2.0, -1.5),  # phi_ef is negative
        ],
    )
    def test_raise_error_when_invalid_values_are_given(self, f_ck: float, lambda_: float, phi_ef: float) -> None:
        """Test invalid values."""
        with pytest.raises((NegativeValueError, LessOrEqualToZeroError)):
            Form5Dot37CreepFactor(f_ck=f_ck, lambda_=lambda_, phi_ef=phi_ef)

    @pytest.mark.parametrize(
        ("representation", "expected"),
        [
            (
                "complete",
                r"K_{\phi} = \max\left(1 + \left(0.35 + \frac{f_{ck}}{200} - \frac{\lambda}{150}\right) \cdot \phi_{ef}; 1\right) "
                r"= \max\left(1 + \left(0.35 + \frac{30.000}{200} - \frac{2.000}{150}\right) \cdot 1.500; 1\right) = 1.730 \ -",
            ),
            ("short", r"K_{\phi} = 1.730 \ -"),
        ],
    )
    def test_latex(self, representation: str, expected: str) -> None:
        """Test the latex representation of the formula."""
        # Example values
        f_ck = 30.0
        lambda_ = 2.0
        phi_ef = 1.5

        # Object to test
        latex = Form5Dot37CreepFactor(f_ck=f_ck, lambda_=lambda_, phi_ef=phi_ef).latex()

        actual = {
            "complete": latex.complete,
            "short": latex.short,
        }

        assert expected == actual[representation], f"{representation} representation failed."



"""Testing formula 12.5 and 12.6 of EN 1992-1-1:2004."""

import pytest

from blueprints.codes.eurocode.en_1992_1_1_2004.chapter_12_plain_and_lightly_reinforced_concrete_structures.formula_12_5_6 import (
    Form12Dot3PlainConcreteShearStress,
    Form12Dot5And6PlainConcreteBendingResistance,
)
from blueprints.validations import LessOrEqualToZeroError


class TestForm12Dot5PlainConcreteBendingResistance:
    """Validation for formula 12.5 from EN 1992-1-1:2004."""

    def test_evaluation(self) -> None:
        """Test the evaluation of the result."""
        # Example values
        f_ctd_pl = 2.5  # MPa
        sigma_cp = 1.0  # MPa
        sigma_c_lim = 1.5  # MPa

        # Object to test
        form_12_5 = Form12Dot5And6PlainConcreteBendingResistance(f_ctd_pl=f_ctd_pl, sigma_cp=sigma_cp, sigma_c_lim=sigma_c_lim)

        # Expected result, manually calculated
        manually_calculated_result = 2.958

        assert round(form_12_5, 3) == pytest.approx(expected=manually_calculated_result, rel=1e-4)

    def test_evaluation_comparison_not_satisfied(self) -> None:
        """Test the evaluation of the result when the comparison is not satisfied."""
        # Example values
        f_ctd_pl = 2.5  # MPa
        sigma_cp = 2.0  # MPa
        sigma_c_lim = 1.5  # MPa

        # Object to test
        form_12_5 = Form12Dot5And6PlainConcreteBendingResistance(f_ctd_pl=f_ctd_pl, sigma_cp=sigma_cp, sigma_c_lim=sigma_c_lim)

        # Expected result, manually calculated
        manually_calculated_result = 3.112

        assert round(form_12_5, 3) == pytest.approx(expected=manually_calculated_result, rel=1e-4)

    @pytest.mark.parametrize(
        ("f_ctd_pl", "sigma_cp"),
        [
            (-2.5, 1.0),
            (2.5, -1.0),
        ],
    )
    def test_raise_error_when_negative_values_are_given(
        self,
        f_ctd_pl: float,
        sigma_cp: float,
    ) -> None:
        """Test negative values for f_ctd_pl and sigma_cp."""
        with pytest.raises(LessOrEqualToZeroError):
            Form12Dot5And6PlainConcreteBendingResistance(f_ctd_pl=f_ctd_pl, sigma_cp=sigma_cp, sigma_c_lim=1.5)

    @pytest.mark.parametrize(
        ("f_ctd_pl", "sigma_cp"),
        [
            (0.0, 1.0),
            (2.5, 0.0),
        ],
    )
    def test_raise_error_when_values_are_less_or_equal_to_zero(
        self,
        f_ctd_pl: float,
        sigma_cp: float,
    ) -> None:
        """Test values less or equal to zero for f_ctd_pl and sigma_cp."""
        with pytest.raises(LessOrEqualToZeroError):
            Form12Dot5And6PlainConcreteBendingResistance(f_ctd_pl=f_ctd_pl, sigma_cp=sigma_cp, sigma_c_lim=1.5)

    @pytest.mark.parametrize(
        ("representation", "expected"),
        [
            (
                "complete",
                r"f_{cvd} = \sqrt{f_{ctd, pl} ^ 2 + \sigma_{cp} \cdot f_{ctd, pl}} = \sqrt{2.500 ^ 2 + 1.000 \cdot 2.500} = 2.958",
            ),
            ("short", r"f_{cvd} = 2.958"),
        ],
    )
    def test_latex(self, representation: str, expected: str) -> None:
        """Test the latex representation of the formula."""
        # Example values
        f_ctd_pl = 2.5  # MPa
        sigma_cp = 1.0  # MPa
        sigma_c_lim = 1.5  # MPa

        # Object to test
        form_12_5_latex = Form12Dot5And6PlainConcreteBendingResistance(
            f_ctd_pl=f_ctd_pl,
            sigma_cp=sigma_cp,
            sigma_c_lim=sigma_c_lim,
        ).latex()

        actual = {
            "complete": form_12_5_latex.complete,
            "short": form_12_5_latex.short,
            "string": str(form_12_5_latex),
        }

        assert actual[representation] == expected, f"{representation} representation failed."

    def test_latex_comparison_not_satisfied(self) -> None:
        """Test the latex representation of the formula when the comparison is not satisfied."""
        # Example values
        f_ctd_pl = 2.5  # MPa
        sigma_cp = 2.0  # MPa
        sigma_c_lim = 1.5  # MPa

        # Object to test
        form_12_5_latex = Form12Dot5And6PlainConcreteBendingResistance(
            f_ctd_pl=f_ctd_pl,
            sigma_cp=sigma_cp,
            sigma_c_lim=sigma_c_lim,
        ).latex()

        expected = (
            r"f_{cvd} = \sqrt{f_{ctd, pl} ^ 2 + \sigma_{cp} \cdot f_{ctd, pl} - "
            r"\left(\frac{\sigma_{cp} - \sigma_{c, lim}}{2}\right) ^ 2} = "
            r"\sqrt{2.500 ^ 2 + 2.000 \cdot 2.500 - \left(\frac{2.000 - "
            r"1.500}2\right) ^ 2} = 3.112"
        )

        assert form_12_5_latex.complete == expected, "Latex representation failed when comparison is not satisfied."

    def test_evaluation_with_sigma_cp_object(self) -> None:
        """Test the evaluation of the result with sigma_cp as an object."""
        # Example values
        f_ctd_pl = 2.5  # MPa
        sigma_cp_object = Form12Dot3PlainConcreteShearStress(n_ed=100000.0, a_cc=50000.0)  # MPa
        sigma_c_lim = 1.5  # MPa

        # Object to test
        form_12_5 = Form12Dot5And6PlainConcreteBendingResistance(f_ctd_pl=f_ctd_pl, sigma_cp=sigma_cp_object, sigma_c_lim=sigma_c_lim)

        manually_calculated_result = 3.112
        assert round(form_12_5, 3) == pytest.approx(expected=round(manually_calculated_result, 3), rel=1e-4)



"""Testing test formula for 3.28, 3.29 and 3.30 of EN 1992-1-1:2004."""

import pytest

from blueprints.codes.eurocode.en_1992_1_1_2004.chapter_3_materials.sub_formula_3_28_29_30 import SubForm3Dot282930Mu


class TestSubForm3Dot282930Mu:
    """Validation for formula for 3.28, 3.29 and 3.30 from EN 1992-1-1:2004."""

    def test_evaluation(self) -> None:
        """Test the evaluation of the result."""
        # Example values
        sigma_pi = 2.4  # MPa
        f_pk = 8.5  # MPa

        sub_form_3_28_29_30 = SubForm3Dot282930Mu(sigma_pi=sigma_pi, f_pk=f_pk)

        # Expected result, manually calculated
        manually_calculated_result = 0.282353

        assert sub_form_3_28_29_30 == pytest.approx(expected=manually_calculated_result, rel=1e-4)

    def test_raise_error_when_negative_f_pk_is_given(self) -> None:
        """Test a negative value."""
        # Example values
        sigma_pi = 2.4  # MPa
        f_pk = -8.5  # MPa

        with pytest.raises(ValueError):
            SubForm3Dot282930Mu(sigma_pi=sigma_pi, f_pk=f_pk)

    @pytest.mark.parametrize(
        ("representation", "expected"),
        [
            (
                "complete",
                r"\mu = \sigma_{pi} / f_{pk} = 2.400 / 8.500 = 0.282",
            ),
            ("short", r"\mu = 0.282"),
        ],
    )
    def test_latex(self, representation: str, expected: str) -> None:
        """Test the latex representation of the formula."""
        # Example values
        sigma_pi = 2.4  # MPa
        f_pk = 8.5  # MPa

        # Object to test
        sub_form_3_28_29_30_latex = SubForm3Dot282930Mu(sigma_pi=sigma_pi, f_pk=f_pk).latex()

        actual = {"complete": sub_form_3_28_29_30_latex.complete, "short": sub_form_3_28_29_30_latex.short}

        assert actual[representation] == expected, f"{representation} representation failed."



"""Testing sub-formula 1 of 3.14 of EN 1992-1-1:2004."""

import pytest

from blueprints.codes.eurocode.en_1992_1_1_2004.chapter_3_materials.formula_3_14 import SubForm3Dot14Eta


class TestSub1Form3Dot14Eta:
    """Validation for sub-formula 1 of 3.14 from EN 1992-1-1:2004."""

    def test_evaluation_1(self) -> None:
        """Test the evaluation of the result."""
        # Example values
        epsilon_c = 0.16  # -
        epsilon_c1 = 0.43  # -

        sub_1_form_3_14 = SubForm3Dot14Eta(epsilon_c=epsilon_c, epsilon_c1=epsilon_c1)

        # Expected result, manually calculated
        manually_calculated_result = 0.37209

        assert sub_1_form_3_14 == pytest.approx(expected=manually_calculated_result, rel=1e-4)

    def test_evaluation_2(self) -> None:
        """Test the evaluation of the result."""
        # Example values
        epsilon_c = 0.3  # -
        epsilon_c1 = 0.2  # -

        sub_1_form_3_14 = SubForm3Dot14Eta(epsilon_c=epsilon_c, epsilon_c1=epsilon_c1)

        # Expected result, manually calculated
        manually_calculated_result = 1.5

        assert sub_1_form_3_14 == pytest.approx(expected=manually_calculated_result, rel=1e-4)

    @pytest.mark.parametrize(
        ("representation", "expected"),
        [
            (
                "complete",
                r"\eta = \epsilon_c / \epsilon_{c1} = 0.160 / 0.430 = 0.372",
            ),
            ("short", r"\eta = 0.372"),
        ],
    )
    def test_latex(self, representation: str, expected: str) -> None:
        """Test the latex representation of the formula."""
        # Example values
        epsilon_c = 0.16  # -
        epsilon_c1 = 0.43  # -

        # Object to test
        form_3_14sub1_latex = SubForm3Dot14Eta(epsilon_c=epsilon_c, epsilon_c1=epsilon_c1).latex()

        actual = {"complete": form_3_14sub1_latex.complete, "short": form_3_14sub1_latex.short}

        assert actual[representation] == expected, f"{representation} representation failed."



"""Testing formula 5.13N of EN 1992-1-1:2004."""

from contextlib import AbstractContextManager
from contextlib import nullcontext as does_not_raise

import pytest

from blueprints.codes.eurocode.en_1992_1_1_2004.chapter_5_structural_analysis.formula_5_13n import (
    Form5Dot13nSlendernessCriterionIsolatedMembers,
    SubForm5Dot13aCreepRatio,
    SubForm5Dot13bMechanicalReinforcementFactor,
    SubForm5Dot13cMomentRatio,
)
from blueprints.validations import EqualToZeroError, LessOrEqualToZeroError, NegativeValueError


class TestForm5Dot13nSlendernessCriterionIsolatedMembers:
    """Validation for formula 5.13N from EN 1992-1-1:2004."""

    def test_evaluation(self) -> None:
        """Tests the evaluation of the result."""
        # Example values
        a = 0.7
        b = 1.1
        c = 0.7
        n_ed = 1_000_000
        a_c = 500_000
        f_cd = 20

        # Object to test
        formula = Form5Dot13nSlendernessCriterionIsolatedMembers(a=a, b=b, c=c, n_ed=n_ed, a_c=a_c, f_cd=f_cd)

        # Expected result, manually calculated
        manually_calculated_result = 34.0893

        assert formula == pytest.approx(expected=manually_calculated_result, rel=1e-4)

    @pytest.mark.parametrize(
        ("a", "b", "c", "n_ed", "a_c", "f_cd"),
        [
            (-0.7, 1.1, 0.7, 1000000.0, 500000.0, 20.0),  # a is negative
            (0.7, -1.1, 0.7, 1000000.0, 500000.0, 20.0),  # b is negative
            (0.7, 1.1, -0.7, 1000000.0, 500000.0, 20.0),  # c is negative
        ],
    )
    def test_raise_error_when_negative_values_are_given(self, a: float, b: float, c: float, n_ed: float, a_c: float, f_cd: float) -> None:
        """Test negative values for a, b, and c."""
        with pytest.raises(NegativeValueError):
            Form5Dot13nSlendernessCriterionIsolatedMembers(a=a, b=b, c=c, n_ed=n_ed, a_c=a_c, f_cd=f_cd)

    @pytest.mark.parametrize(
        ("a", "b", "c", "n_ed", "a_c", "f_cd"),
        [
            (0.7, 1.1, 0.7, 0.0, 500000.0, 20.0),  # n_ed is zero
            (0.7, 1.1, 0.7, -1000000.0, 500000.0, 20.0),  # n_ed is negative
            (0.7, 1.1, 0.7, 1000000.0, 0.0, 20.0),  # a_c is zero
            (0.7, 1.1, 0.7, 1000000.0, -500000.0, 20.0),  # a_c is negative
            (0.7, 1.1, 0.7, 1000000.0, 500000.0, 0.0),  # f_cd is zero
            (0.7, 1.1, 0.7, 1000000.0, 500000.0, -20.0),  # f_cd is negative
        ],
    )
    def test_raise_error_when_zero_or_negative_values_are_given(self, a: float, b: float, c: float, n_ed: float, a_c: float, f_cd: float) -> None:
        """Test zero and negative values for n_ed, a_c, and f_cd."""
        with pytest.raises(LessOrEqualToZeroError):
            Form5Dot13nSlendernessCriterionIsolatedMembers(a=a, b=b, c=c, n_ed=n_ed, a_c=a_c, f_cd=f_cd)

    @pytest.mark.parametrize(
        ("representation", "expected"),
        [
            (
                "complete",
                r"\lambda_{lim} = \frac{20 \cdot A \cdot B \cdot C}{\sqrt{N_{Ed} \cdot A_c \cdot f_{cd}}} = "
                r"\frac{20 \cdot 0.700 \cdot 1.100 \cdot 0.700}{\sqrt{1000000.000 \cdot 500000.000 \cdot 20.000}} "
                r"= 34.089 \ -",
            ),
            (
                "complete_with_units",
                r"\lambda_{lim} = \frac{20 \cdot A \cdot B \cdot C}{\sqrt{N_{Ed} \cdot A_c \cdot f_{cd}}} = "
                r"\frac{20 \cdot 0.700 \cdot 1.100 \cdot 0.700}{\sqrt{1000000.000 \ N \cdot 500000.000 "
                r"\ mm^2 \cdot 20.000 \ MPa}} = 34.089 \ -",
            ),
            ("short", r"\lambda_{lim} = 34.089 \ -"),
        ],
    )
    def test_latex(self, representation: str, expected: str) -> None:
        """Test the latex representation of the formula."""
        # Example values
        a = 0.7
        b = 1.1
        c = 0.7
        n_ed = 1000000.0
        a_c = 500000.0
        f_cd = 20.0

        # Object to test
        latex = Form5Dot13nSlendernessCriterionIsolatedMembers(a=a, b=b, c=c, n_ed=n_ed, a_c=a_c, f_cd=f_cd).latex()

        actual = {
            "complete": latex.complete,
            "complete_with_units": latex.complete_with_units,
            "short": latex.short,
        }

        assert expected == actual[representation], f"{representation} representation failed."


class TestForm5Dot13aCreepFactor:
    """Validation for formula 5.13a from EN 1992-1-1:2004."""

    def test_evaluation(self) -> None:
        """Tests the evaluation of the result."""
        # Example value
        phi_ef = 2.5

        # Object to test
        formula = SubForm5Dot13aCreepRatio(phi_ef=phi_ef)

        # Expected result, manually calculated
        manually_calculated_result = 0.6666666667

        assert formula == pytest.approx(expected=manually_calculated_result, rel=1e-4)

    def test_raise_error_when_invalid_values_are_given(self) -> None:
        """Test invalid values."""
        with pytest.raises(NegativeValueError):
            SubForm5Dot13aCreepRatio(phi_ef=-1.0)

    @pytest.mark.parametrize(
        ("representation", "expected"),
        [
            (
                "complete",
                r"A = \frac{1}{(1 + 0.2 \cdot \phi_{ef})} = \frac{1}{(1 + 0.2 \cdot 2.500)} = 0.667 \ -",
            ),
            ("short", r"A = 0.667 \ -"),
        ],
    )
    def test_latex(self, representation: str, expected: str) -> None:
        """Test the latex representation of the formula."""
        # Example value
        phi_ef = 2.5

        # Object to test
        latex = SubForm5Dot13aCreepRatio(phi_ef=phi_ef).latex()

        actual = {
            "complete": latex.complete,
            "complete_with_units": latex.complete_with_units,
            "short": latex.short,
        }

        assert expected == actual[representation], f"{representation} representation failed."


class TestForm5Dot13bMechanicalReinforcementFactor:
    """Validation for formula 5.13b from EN 1992-1-1:2004."""

    def test_evaluation(self) -> None:
        """Tests the evaluation of the result."""
        # Example values
        a_s = 1000.0
        f_yd = 500.0
        a_c = 10000.0
        f_cd = 20.0

        # Object to test
        formula = SubForm5Dot13bMechanicalReinforcementFactor(a_s=a_s, f_yd=f_yd, a_c=a_c, f_cd=f_cd)

        # Expected result, manually calculated
        manually_calculated_result = 2.4494  # dimensionless

        assert formula == pytest.approx(expected=manually_calculated_result, rel=1e-4)

    @pytest.mark.parametrize(
        ("a_s", "f_yd", "a_c", "f_cd"),
        [
            (-1000.0, 500.0, 10000.0, 20.0),  # a_s is negative
            (1000.0, -500.0, 10000.0, 20.0),  # f_yd is negative
        ],
    )
    def test_raise_error_when_negative_values_are_given(self, a_s: float, f_yd: float, a_c: float, f_cd: float) -> None:
        """Test negative values."""
        with pytest.raises(NegativeValueError):
            SubForm5Dot13bMechanicalReinforcementFactor(a_s=a_s, f_yd=f_yd, a_c=a_c, f_cd=f_cd)

    @pytest.mark.parametrize(
        ("a_s", "f_yd", "a_c", "f_cd"),
        [
            (1000.0, 500.0, 0.0, 20.0),  # a_c is zero
            (1000.0, 500.0, -10000.0, 20.0),  # a_c is negative
            (1000.0, 500.0, 10000.0, 0.0),  # f_cd is zero
            (1000.0, 500.0, 10000.0, -20.0),  # f_cd is negative
        ],
    )
    def test_raise_error_when_less_or_equal_to_zero_values_are_given(self, a_s: float, f_yd: float, a_c: float, f_cd: float) -> None:
        """Test less or equal to zero values."""
        with pytest.raises(LessOrEqualToZeroError):
            SubForm5Dot13bMechanicalReinforcementFactor(a_s=a_s, f_yd=f_yd, a_c=a_c, f_cd=f_cd)

    @pytest.mark.parametrize(
        ("representation", "expected"),
        [
            (
                "complete",
                r"B = \sqrt{1 + 2 \cdot \frac{A_s \cdot f_{yd}}{A_c \cdot f_{cd}}} = \sqrt{1 + 2 \cdot "
                r"\frac{1000.000 \cdot 500.000}{10000.000 \cdot 20.000}} = 2.449 \ -",
            ),
            (
                "complete_with_units",
                r"B = \sqrt{1 + 2 \cdot \frac{A_s \cdot f_{yd}}{A_c \cdot f_{cd}}} = \sqrt{1 + 2 \cdot "
                r"\frac{1000.000 \ mm^2 \cdot 500.000 \ MPa}{10000.000 \ mm^2 \cdot 20.000 \ MPa}} = 2.449 \ -",
            ),
            ("short", r"B = 2.449 \ -"),
        ],
    )
    def test_latex(self, representation: str, expected: str) -> None:
        """Test the latex representation of the formula."""
        # Example values
        a_s = 1000.0
        f_yd = 500.0
        a_c = 10000.0
        f_cd = 20.0

        # Object to test
        latex = SubForm5Dot13bMechanicalReinforcementFactor(a_s=a_s, f_yd=f_yd, a_c=a_c, f_cd=f_cd).latex()

        actual = {
            "complete": latex.complete,
            "complete_with_units": latex.complete_with_units,
            "short": latex.short,
        }

        assert expected == actual[representation], f"{representation} representation failed."


class TestForm5Dot13cMomentFactor:
    """Validation for formula 5.13c from EN 1992-1-1:2004."""

    def test_evaluation(self) -> None:
        """Tests the evaluation of the result."""
        # Example values
        m_01 = 100.0
        m_02 = 150.0

        # Object to test
        formula = SubForm5Dot13cMomentRatio(m_01=m_01, m_02=m_02)

        # Expected result, manually calculated
        manually_calculated_result = 1.033333  # dimensionless

        assert formula == pytest.approx(expected=manually_calculated_result, rel=1e-4)

    @pytest.mark.parametrize(
        ("m_01", "m_02", "expectation"),
        [
            (100.0, 150.0, does_not_raise()),
            (70, 70, does_not_raise()),  # m_01 == m_02
            (150, 0, pytest.raises(EqualToZeroError)),  # m_02 is zero
            (100, 50, pytest.raises(ValueError)),  # m_02 > m_01
        ],
        ids=[
            "passes",
            "m_01==m_02",
            "m_02=0",
            "m_02>m_01",
        ],
    )
    def test_raise_error_incorrect_args(self, m_01: float, m_02: float, expectation: AbstractContextManager) -> None:
        """Test if errors are raised."""
        with expectation:
            assert SubForm5Dot13cMomentRatio(m_01=m_01, m_02=m_02) is not None

    @pytest.mark.parametrize(
        ("representation", "expected"),
        [
            (
                "complete",
                r"C = 1.7 - \frac{M_{01}}{M_{02}} = 1.7 - \frac{100.000}{150.000} = 1.033 \ -",
            ),
            (
                "complete_with_units",
                r"C = 1.7 - \frac{M_{01}}{M_{02}} = 1.7 - \frac{100.000 \ kNm}{150.000 \ kNm} = 1.033 \ -",
            ),
            ("short", r"C = 1.033 \ -"),
        ],
    )
    def test_latex(self, representation: str, expected: str) -> None:
        """Test the latex representation of the formula."""
        # Example values
        m_01 = 100.0
        m_02 = 150.0

        # Object to test
        latex = SubForm5Dot13cMomentRatio(m_01=m_01, m_02=m_02).latex()

        actual = {
            "complete": latex.complete,
            "complete_with_units": latex.complete_with_units,
            "short": latex.short,
        }

        assert expected == actual[representation], f"{representation} representation failed."



"""Formula 9.12N from EN 1992-1-1:2004: Chapter 9 - Detailling and specific rules."""

from blueprints.codes.eurocode.en_1992_1_1_2004 import EN_1992_1_1_2004
from blueprints.codes.formula import Formula
from blueprints.codes.latex_formula import LatexFormula
from blueprints.type_alias import KN, MM2, MPA
from blueprints.unit_conversion import KN_TO_N
from blueprints.validations import raise_if_negative


class Form9Dot12nMinimumLongitudinalReinforcementColumns(Formula):
    """Class representing the formula 9.12N for the calculation of the minimum longitudinal reinforcement for columns."""

    label = "9.12N"
    source_document = EN_1992_1_1_2004

    def __init__(
        self,
        n_ed: KN,
        f_yd: MPA,
        a_c: MM2,
    ) -> None:
        r"""[$A_{s,min}$] Minimum longitudinal reinforcement for columns [$mm^2$].

        EN 1992-1-1:2004 art.9.5.2(2) - Formula (9.12N)

        Parameters
        ----------
        n_ed: KN
            [$N_{Ed}$] Design value of axial force [$kN$].
        f_yd: MPA
            [$f_{yd}$] Design yield strength reinforcement steel [$MPa$].
        a_c: MM2
            [$A_c$] Concrete cross-sectional area [$mm^2$].
        """
        super().__init__()
        self.n_ed = n_ed
        self.f_yd = f_yd
        self.a_c = a_c

    @staticmethod
    def _evaluate(
        n_ed: KN,
        f_yd: MPA,
        a_c: MM2,
    ) -> MM2:
        """For more detailed documentation see the class docstring."""
        raise_if_negative(n_ed=n_ed, f_yd=f_yd, a_c=a_c)
        return max(0.1 * n_ed * KN_TO_N / f_yd, 0.002 * a_c)

    def latex(self, n: int = 2) -> LatexFormula:
        """Returns LatexFormula object for formula 9.12N."""
        return LatexFormula(
            return_symbol=r"A_{s,min}",
            result=f"{self:.{n}f}",
            equation=r"\max( \frac{0.10 \cdot N_{Ed}}{f_{yd}}, 0.002 \cdot A_c )",
            numeric_equation=rf"\max( \frac{{0.10 \cdot {self.n_ed:.{n}f}}}{{{self.f_yd:.{n}f}}}, 0.002 \cdot {self.a_c:.{n}f} )",
            comparison_operator_label="=",
        )
